# Copyright 2016 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Database-related models and functions."""

from collections import namedtuple

import psycopg2
import sqlalchemy
from sqlalchemy import sql
from sqlalchemy.exc import IntegrityError
import sqlite3

from gpgservice.config import config


def create_engine():
    if config.get('database', 'type').lower().strip() == 'postgres':
        def connect_to_db():
            return psycopg2.connect(
                host=config.get('database', 'host'),
                dbname=config.get('database', 'name'),
                user=config.get('database', 'user'),
                password=config.get('database', 'password'))
        return sqlalchemy.create_engine('postgresql://', creator=connect_to_db)
    else:
        # Create an in-memory sqlite database.
        return sqlalchemy.create_engine('sqlite://')


engine = create_engine()
metadata = sqlalchemy.MetaData(bind=engine)


GPGKeyTable = sqlalchemy.Table(
    'GPGKey',
    metadata,
    sqlalchemy.Column('owner', sqlalchemy.String, nullable=False),
    sqlalchemy.Column('keyid', sqlalchemy.String, nullable=False),
    sqlalchemy.Column('fingerprint', sqlalchemy.String, primary_key=True,
                      nullable=False),
    sqlalchemy.Column('keysize', sqlalchemy.Integer, nullable=False),
    sqlalchemy.Column('algorithm',
                      sqlalchemy.Enum('R', 'g', 'D', 'G', name='algorithm'),
                      nullable=False),
    sqlalchemy.Column('enabled', sqlalchemy.Boolean, nullable=False),
    sqlalchemy.Column('can_encrypt', sqlalchemy.Boolean, nullable=False),
)


metadata.create_all(engine)

GPGKey = namedtuple('GPGKey',
                    ['owner', 'keyid', 'fingerprint', 'keysize', 'algorithm',
                     'enabled', 'can_encrypt'])


def key_exists(connection, fingerprint):
    query = sql.select(
        [GPGKeyTable.c.fingerprint]
    ).where(
        GPGKeyTable.c.fingerprint == fingerprint
    )
    return connection.execute(query).scalar() is not None


class DuplicateKeyInsertionError(ValueError):
    pass


def insert_key(connection, key):
    """Insert a key into the database.

    :param key: a GPGKey instance.
    :raises DuplicateKeyInsertionError: if the key to be inserted has a
        duplicate fingerprint to one already in the database.

    """
    query = GPGKeyTable.insert().values(
        owner=key.owner,
        keyid=key.keyid,
        fingerprint=key.fingerprint,
        keysize=key.keysize,
        algorithm=key.algorithm,
        enabled=key.enabled,
        can_encrypt=key.can_encrypt,
    )
    try:
        connection.execute(query)
    except IntegrityError as e:
        # Ensure we only raise DuplicateKeyInsertionError when we're sure
        # the IntegrityError was caused by the unique constraint failing.
        # Postgres gives us a code, sqlite gives us a string.
        if isinstance(e.orig, psycopg2.IntegrityError):
            if e.orig.pgcode != '23505':
                raise
        elif isinstance(e.orig, sqlite3.IntegrityError):
            if not e.orig.message.startswith("UNIQUE constraint failed:"):
                raise
        raise DuplicateKeyInsertionError(
            "Fingerprint %r already in database" % key.fingerprint)


def get_keys_for_owner(connection, owner):
    """Get all the keys for a particular owner.

    Returns (id, fingerprint, keysize, algorithm, enabled, can_encrypt) tuples

    """
    query = sql.select([GPGKeyTable]).where(GPGKeyTable.c.owner == owner)
    result = connection.execute(query)
    for row in result:
        yield GPGKey(*row)


def owner_owns_fingerprint(connection, owner, fingerprint):
    """Return True if `owner` is the owner of `fingerprint`.

    Will return false if either the fingerprint is not in the database at all,
    or if it is in the database, but registerred to a different owner.

    """
    query = sql.select(
        [GPGKeyTable.c.fingerprint]
    ).where(
        GPGKeyTable.c.fingerprint == fingerprint
    ).where(
        GPGKeyTable.c.owner == owner
    )
    return connection.execute(query).scalar() is not None


def disable_key(connection, fingerprint):
    """Set the 'enabled' columns for a key to False.

    Returns True if anything was modified, False otherwise.
    """
    return _set_key_enabled(connection, fingerprint, False)


def enable_key(connection, fingerprint):
    """Set the 'enabled' columns for a key to True.

    Returns True if anything was modified, False otherwise.
    """
    return _set_key_enabled(connection, fingerprint, True)


def _set_key_enabled(connection, fingerprint, enabled):
    query = GPGKeyTable.update().where(
        GPGKeyTable.c.fingerprint == fingerprint).values(
        enabled=enabled)
    return connection.execute(query).rowcount > 0


def get_keys_by_fingerprints(connection, fingerprint):
    if not isinstance(fingerprint, (list, tuple)):
        fingerprint = [fingerprint]
    query = sql.select([GPGKeyTable]).where(
                GPGKeyTable.c.fingerprint.in_(fingerprint))
    result = connection.execute(query)
    return [GPGKey(*k) for k in result]
