# Copyright 2009-2016 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__all__ = [
    'GPGKeyCannotEncrypt',
    'GPGKeyDoesNotExistOnServer',
    'GPGKeyExpired',
    'GPGKeyNotFoundError',
    'GPGKeyRevoked',
    'GPGKeyTemporarilyNotFoundError',
    'GPGUploadFailure',
    'GPGVerificationError',
    'MoreThanOneGPGKeyFound',
    'SecretGPGKeyImportDetected',
    ]


class MoreThanOneGPGKeyFound(Exception):
    """More than one GPG key was found.

    And we don't know which one to import.
    """


class GPGKeyNotFoundError(Exception):
    """The GPG key with the given fingerprint was not found on the keyserver.
    """

    def __init__(self, fingerprint, message=None):
        self.fingerprint = fingerprint
        if message is None:
            message = (
                "No GPG key found with the given content: %s" % (fingerprint))
        super(GPGKeyNotFoundError, self).__init__(message)


class GPGKeyTemporarilyNotFoundError(GPGKeyNotFoundError):
    """The GPG key with the given fingerprint was not found on the keyserver.

    The reason is a timeout while accessing the server, a general
    server error, a network problem or some other temporary issue.
    """
    def __init__(self, fingerprint):
        message = (
            "GPG key %s not found due to a server or network failure."
            % fingerprint)
        super(GPGKeyTemporarilyNotFoundError, self).__init__(
            fingerprint, message)


class GPGKeyDoesNotExistOnServer(GPGKeyNotFoundError):
    """The GPG key with the given fingerprint was not found on the keyserver.

    The server returned an explicit "not found".
    """
    def __init__(self, fingerprint):
        message = (
            "GPG key %s does not exist on the keyserver." % fingerprint)
        super(GPGKeyDoesNotExistOnServer, self).__init__(
            fingerprint, message)


class GPGKeyRevoked(Exception):
    """The given GPG key was revoked."""

    def __init__(self, key):
        self.key = key
        super(GPGKeyRevoked, self).__init__(
            "%s has been publicly revoked" % (key.keyid, ))


class GPGKeyExpired(Exception):
    """The given GPG key has expired."""

    def __init__(self, key):
        self.key = key
        super(GPGKeyExpired, self).__init__("%s has expired" % (key.keyid, ))


class SecretGPGKeyImportDetected(Exception):
    """An attempt to import a secret GPG key."""


class GPGUploadFailure(Exception):
    """Raised when a key upload failed.

    Typically when a keyserver is not reachable.
    """


class GPGVerificationError(Exception):
    """OpenPGP verification error."""


class GPGKeyCannotEncrypt(Exception):
    """A sign-only key was used for an encryption operation."""
