# Copyright 2009-2016 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import atexit
from collections import namedtuple
import httplib
import os
import shutil
import socket
from StringIO import StringIO
import subprocess
import tempfile
import urllib
import urllib2

import gpgme

from gpgservice.config import config
from gpgservice.errors import (
    GPGKeyCannotEncrypt,
    GPGKeyDoesNotExistOnServer,
    GPGKeyExpired,
    GPGKeyNotFoundError,
    GPGKeyRevoked,
    GPGKeyTemporarilyNotFoundError,
    GPGUploadFailure,
    GPGVerificationError,
    MoreThanOneGPGKeyFound,
    SecretGPGKeyImportDetected,
    )
from gpgservice.validators import (
    valid_email,
    valid_fingerprint,
    )
# TODO: Either import these utilities, or find an alternative in the stdlib
# from lp.services.timeout import (
#    TimeoutError,
#    urlfetch,
#    )
# from lp.services.webapp import errorlog

__metaclass__ = type

__all__ = [
    'GPGHandler',
    'GPGKeyAlgorithm',
    'PymeKey',
    'PymeSignature',
    'PymeUserId',
    'handler_singleton',
]

signing_only_param = """
<GnupgKeyParms format="internal">
  Key-Type: RSA
  Key-Usage: sign
  Key-Length: 4096
  Name-Real: %(name)s
  Expire-Date: 0
</GnupgKeyParms>
"""


class GPGKeyAlgorithm:
    """
    GPG Compliant Key Algorithms Types:

    1 : "R", # RSA
    16: "g", # ElGamal
    17: "D", # DSA
    20: "G", # ElGamal, compromised

    This class is used to mat between the integer values (used by pgpme), the
    'name' (which is a single character identifier) and the title. This was
    originally a lazr.enum class, but we've dropped the zope dependency. This
    class is made to mimic the API of the lazr enum, which is why it's a little
    odd.

    """
    Algorithm = namedtuple('Algorithm', ['name', 'title'])

    R = Algorithm('R', 'R')
    LITTLE_G = Algorithm('LITTLE_G', 'g')
    D = Algorithm('D', 'D')
    G = Algorithm('G', 'G')

    items = {
        1: R,
        'R': R,
        16: LITTLE_G,
        'LITTLE_G': LITTLE_G,
        17: D,
        'D': D,
        20: G,
        'G': G
    }


class GPGHandler:
    """Handler to perform OpenPGP operations."""

    def __init__(self):
        """Initialize environment variable."""
        self._setNewHome()
        os.environ['GNUPGHOME'] = self.home

    def _setNewHome(self):
        """Create a new directory containing the required configuration.

        This method is called inside the class constructor and genereates
        a new directory (name randomly generated with the 'gpg-' prefix)
        containing the proper file configuration and options.

        Also installs an atexit handler to remove the directory on normal
        process termination.
        """
        self.home = tempfile.mkdtemp(prefix='gpg-')
        confpath = os.path.join(self.home, 'gpg.conf')
        conf = open(confpath, 'w')
        # set needed GPG options, 'auto-key-retrieve' is necessary for
        # automatically retrieve from the keyserver unknown key when
        # verifying signatures and 'no-auto-check-trustdb' avoid wasting
        # time verifying the local keyring consistence.
        conf.write(
            'keyserver hkp://%s\n'
            'keyserver-options auto-key-retrieve\n'
            'no-auto-check-trustdb\n' % config.get('gpghandler', 'host'))
        conf.close()
        # create a local atexit handler to remove the configuration directory
        # on normal termination.

        atexit.register(self.removeHome)

    def removeHome(self):
        """Remove GNUPGHOME directory.

        This is called via an `atexit` handler, and should not be called during
        normal operation.
        """
        if os.path.exists(self.home):
            shutil.rmtree(self.home)

    def sanitizeFingerprint(self, fingerprint):
        """Return sanitized fingerprint if well-formed.

        If the firgerprint cannot be sanitized return None.
        """
        # remove whitespaces, truncate to max of 40 (as per v4 keys) and
        # convert to upper case
        fingerprint = fingerprint.replace(' ', '')
        fingerprint = fingerprint[:40].upper()

        if not valid_fingerprint(fingerprint):
            return None

        return fingerprint

    def resetLocalState(self):
        """Reset the local state.

        Resets OpenPGP keyrings and trust database.
        """
        # remove the public keyring, private keyring and the trust DB
        for filename in ['pubring.gpg', 'secring.gpg', 'trustdb.gpg']:
            filename = os.path.join(self.home, filename)
            if os.path.exists(filename):
                os.remove(filename)

    def verifySignature(self, content, signature=None):
        """See `getVerifiedSignature`.

        Suppress all exceptions and simply return None if the could not
        be verified.
        """
        try:
            return self.getVerifiedSignature(content, signature)
        except GPGVerificationError:
            # Swallow GPG Verification Errors
            pass
        return None

    def getVerifiedSignatureResilient(self, content, signature=None):
        """Wrapper for getVerifiedSignature.

        It calls the target method exactly 3 times.

        Return the result if it succeed during the cycle, otherwise
        capture the errors and emits at the end GPGVerificationError
        with the stored error information.
        """
        errors = []

        for i in range(3):
            try:
                signature = self.getVerifiedSignature(content, signature)
            except GPGVerificationError as info:
                errors.append(info)
            else:
                return signature

        stored_errors = [str(err) for err in errors]

        raise GPGVerificationError(
            "Verification failed 3 times: %s " % stored_errors)

    def getVerifiedSignature(self, content, signature=None):
        """Returns a PymeSignature object if content is correctly signed.

        If signature is None, we assume content is clearsigned. Otherwise
        it stores the detached signature and content should contain the
        plain text in question.

        content and signature must be 8-bit encoded str objects. It's up to
        the caller to encode or decode as appropriate.

        The only exception likely to be propogated out is GPGVerificationError

        :param content: The content to be verified as string;
        :param signature: The signature as string (or None if content is
            clearsigned)

        :raise GPGVerificationError: if the signature cannot be verified.
        :return: a `PymeSignature` object.
        """

        assert not isinstance(content, unicode)
        assert not isinstance(signature, unicode)

        ctx = gpgme.Context()

        # from `info gpgme` about gpgme_op_verify(SIG, SIGNED_TEXT, PLAIN):
        #
        # If SIG is a detached signature, then the signed text should be
        # provided in SIGNED_TEXT and PLAIN should be a null pointer.
        # Otherwise, if SIG is a normal (or cleartext) signature,
        # SIGNED_TEXT should be a null pointer and PLAIN should be a
        # writable data object that will contain the plaintext after
        # successful verification.

        if signature:
            # store detach-sig
            sig = StringIO(signature)
            # store the content
            plain = StringIO(content)
            args = (sig, plain, None)
        else:
            # store clearsigned signature
            sig = StringIO(content)
            # writeable content
            plain = StringIO()
            args = (sig, None, plain)

        # process it
        try:
            signatures = ctx.verify(*args)
        except gpgme.GpgmeError as e:
            error = GPGVerificationError(e.strerror)
            for attr in ("args", "code", "signatures", "source"):
                if hasattr(e, attr):
                    value = getattr(e, attr)
                    setattr(error, attr, value)
            raise error

        # XXX jamesh 2006-01-31:
        # We raise an exception if we don't get exactly one signature.
        # If we are verifying a clear signed document, multiple signatures
        # may indicate two differently signed sections concatenated
        # together.
        # Multiple signatures for the same signed block of data is possible,
        # but uncommon.  If people complain, we'll need to examine the issue
        # again.

        # if no signatures were found, raise an error:
        if len(signatures) == 0:
            raise GPGVerificationError('No signatures found')
        # we only expect a single signature:
        if len(signatures) > 1:
            raise GPGVerificationError('Single signature expected, '
                                       'found multiple signatures')

        signature = signatures[0]
        # signature.status == 0 means "Ok"
        if signature.status is not None:
            error = GPGVerificationError(signature.status.strerror)
            for attr in ("args", "code", "source"):
                if hasattr(signature.status, attr):
                    value = getattr(signature.status, attr)
                    setattr(error, attr, value)
            raise error

        # supporting subkeys by retriving the full key from the
        # keyserver and use the master key fingerprint.
        try:
            key = self.retrieveKey(signature.fpr)
        except GPGKeyNotFoundError:
            raise GPGVerificationError(
                "Unable to map subkey: %s" % signature.fpr)

        # return the signature container
        return PymeSignature(
            fingerprint=key.fingerprint,
            plain_data=plain.getvalue(),
            timestamp=signature.timestamp)

    def importPublicKey(self, content):
        """Import the given public key into our local keyring.

        If the secret key's ASCII armored content is given,
        SecretGPGKeyDetected is raised.

        If no key is found, GPGKeyNotFoundError is raised.  On the other
        hand, if more than one key is found, MoreThanOneGPGKeyFound is
        raised.

        :param content: public key ASCII armored content (must be an ASCII
            string (it's up to the caller to encode or decode properly);
        :return: a `PymeKey` object referring to the public key imported.
        """
        assert isinstance(content, str)
        context = gpgme.Context()
        context.armor = True

        newkey = StringIO(content)
        result = context.import_(newkey)

        if len(result.imports) == 0:
            raise GPGKeyNotFoundError(content)

        # Check the status of all imported keys to see if any of them is
        # a secret key.  We can't rely on result.secret_imported here
        # because if there's a secret key which is already imported,
        # result.secret_imported will be 0.
        for fingerprint, res, status in result.imports:
            if status & gpgme.IMPORT_SECRET != 0:
                raise SecretGPGKeyImportDetected(
                    "GPG key '%s' is a secret key." % fingerprint)

        if len(result.imports) > 1:
            raise MoreThanOneGPGKeyFound('Found %d GPG keys when importing %s'
                                         % (len(result.imports), content))

        fingerprint, res, status = result.imports[0]
        key = PymeKey(fingerprint)
        assert key.exists_in_local_keyring
        return key

    def importSecretKey(self, content):
        """Import the given secret key into our local keyring.

        If no key is found, GPGKeyNotFoundError is raised.  On the other
        hand, if more than one key is found, MoreThanOneGPGKeyFound is
        raised.

        :param content: secret key ASCII armored content (must be an ASCII
            string (it's up to the caller to encode or decode properly);
        :return: a `PymeKey` object referring to the secret key imported.
        """
        assert isinstance(content, str)

        # Make sure that gpg-agent doesn't interfere.
        if 'GPG_AGENT_INFO' in os.environ:
            del os.environ['GPG_AGENT_INFO']

        context = gpgme.Context()
        context.armor = True
        newkey = StringIO(content)
        import_result = context.import_(newkey)

        secret_imports = [
            fingerprint
            for fingerprint, result, status in import_result.imports
            if status & gpgme.IMPORT_SECRET]
        if len(secret_imports) == 0:
            raise GPGKeyNotFoundError(content)
        if len(secret_imports) != 1:
            raise MoreThanOneGPGKeyFound(
                'Found %d secret GPG keys when importing %s'
                % (len(secret_imports), content))

        fingerprint, result, status = import_result.imports[0]
        try:
            key = context.get_key(fingerprint, True)
        except gpgme.GpgmeError:
            return None

        key = PymeKey.newFromGpgmeKey(key)
        assert key.exists_in_local_keyring
        return key

    def generateKey(self, name):
        """Generate a new GPG key with the given name.

        Currently only passwordless, signo-only 1024-bit RSA keys are
        generated.

        :param name: unicode to be included in the key paramenters, 'comment'
            and 'email' will be empty. It's content will be encoded to
            'utf-8' internally.
        :raise AssertionError: if the key generation is not exaclty what
            we expect.

        :return: a `PymeKey` object for the just-generated secret key.
        """
        context = gpgme.Context()

        # Make sure that gpg-agent doesn't interfere.
        if 'GPG_AGENT_INFO' in os.environ:
            del os.environ['GPG_AGENT_INFO']

        # Only 'utf-8' encoding is supported by gpgme.
        # See more information at:
        # http://pyme.sourceforge.net/doc/gpgme/Generating-Keys.html
        result = context.genkey(
            signing_only_param % {'name': name.encode('utf-8')})

        # Right, it might seem paranoid to have this many assertions,
        # but we have to take key generation very seriously.
        assert result.primary, 'Secret key generation failed.'
        assert not result.sub, (
            'Only sign-only RSA keys are safe to be generated')

        secret_keys = list(self.localKeys(result.fpr, secret=True))

        assert len(secret_keys) == 1, 'Found %d secret GPG keys for %s' % (
            len(secret_keys), result.fpr)

        key = secret_keys[0]

        assert key.fingerprint == result.fpr, (
            'The key in the local keyring does not match the one generated.')
        assert key.exists_in_local_keyring, (
            'The key does not seem to exist in the local keyring.')

        return key

    def encryptContent(self, content, fingerprint):
        """Encrypt the given content for the given fingerprint.

        content must be a traditional string. It's up to the caller to
        encode or decode properly. Fingerprint must be hexadecimal string.

        :param content: the content to be encrypted, as a byte string.
        :param fingerprint: the OpenPGP key's fingerprint.

        :return: the encrypted content or None if failed.
        """
        if isinstance(content, unicode):
            raise TypeError('Content cannot be Unicode.')

        # setup context
        ctx = gpgme.Context()
        ctx.armor = True

        # setup containers
        plain = StringIO(content)
        cipher = StringIO()

        # retrive gpgme key object
        try:
            key = ctx.get_key(fingerprint.encode('ascii'), 0)
        except gpgme.GpgmeError:
            return None

        if not key.can_encrypt:
            raise GPGKeyCannotEncrypt('key %s can not be used for encryption'
                                      % fingerprint)

        # encrypt content
        ctx.encrypt([key], gpgme.ENCRYPT_ALWAYS_TRUST, plain, cipher)

        return cipher.getvalue()

    def signContent(self, content, key_fingerprint, password='', mode=None):
        """Signs content with a given GPG fingerprint.

        :param content: the content to sign.
        :param key_fingerprint: the fingerprint of the key to use when
            signing the content.
        :param password: optional password to the key identified by
            key_fingerprint, the default value is '',
        :param mode: optional he type of GPG signature to produce, the
            default mode is gpgme.SIG_MODE_CLEAR (clearsigned signatures)

        :return: The ASCII-armored signature for the content.
        """
        if not isinstance(content, str):
            raise TypeError('Content should be a string.')

        if mode is None:
            mode = gpgme.SIG_MODE_CLEAR

        # Find the key and make it the only one allowed to sign content
        # during this session.
        context = gpgme.Context()
        context.armor = True

        key = context.get_key(key_fingerprint.encode('ascii'), True)
        context.signers = [key]

        # Set up containers.
        plaintext = StringIO(content)
        signature = StringIO()

        # Make sure that gpg-agent doesn't interfere.
        if 'GPG_AGENT_INFO' in os.environ:
            del os.environ['GPG_AGENT_INFO']

        def passphrase_cb(uid_hint, passphrase_info, prev_was_bad, fd):
            os.write(fd, '%s\n' % password)
        context.passphrase_cb = passphrase_cb

        # Sign the text.
        try:
            context.sign(plaintext, signature, mode)
        except gpgme.GpgmeError:
            return None

        return signature.getvalue()

    def localKeys(self, filter=None, secret=False):
        """Get an iterator of the keys this gpg handler
        already knows about.
        """
        ctx = gpgme.Context()

        for key in ctx.keylist(filter, secret):
            yield PymeKey.newFromGpgmeKey(key)

    def retrieveKey(self, fingerprint):
        """Retrieve the key information from the local keyring.

        If the key with the given fingerprint is not present in the local
        keyring, first import it from the key server into the local keyring.

        :param fingerprint: The key fingerprint, which must be an hexadecimal
            string.
        :raise GPGKeyNotFoundError: if the key is not found neither in the
            local keyring nor in the key server.
        :return: a `PymeKey`object containing the key information.
        """
        # XXX cprov 2005-07-05:
        # Integrate it with the furure proposal related
        # synchronization of the local key ring with the
        # global one. It should basically consists of be
        # aware of a revoked flag coming from the global
        # key ring, but it needs "specing"
        key = PymeKey(fingerprint.encode('ascii'))
        if not key.exists_in_local_keyring:
            pubkey = self._getPubKey(fingerprint)
            key = self.importPublicKey(pubkey)
        return key

    def retrieveActiveKey(self, fingerprint):
        """Retrieve key information, raise errors if the key is not active.

        Exactly like `retrieveKey` except raises errors if the key is expired
        or has been revoked.

        :param fingerprint: The key fingerprint, which must be an hexadecimal
            string.
        :raise GPGKeyNotFoundError: if the key is not found neither in the
            local keyring nor in the key server.
        :return: a `PymeKey`object containing the key information.
        """
        key = self.retrieveKey(fingerprint)
        if key.revoked:
            raise GPGKeyRevoked(key)
        if key.expired:
            raise GPGKeyExpired(key)
        return key

    def _submitKey(self, content):
        """Submit an ASCII-armored public key export to the keyserver.

        It issues a POST at /pks/add on the keyserver specified in the
        configuration.
        """
        keyserver_http_url = '%s:%s' % (
            config.get('gpghandler', 'host'),
            config.getint('gpghandler', 'port'))

        conn = httplib.HTTPConnection(keyserver_http_url)
        params = urllib.urlencode({'keytext': content})
        headers = {
            "Content-type": "application/x-www-form-urlencoded",
            "Accept": "text/plain",
            }

        try:
            conn.request("POST", "/pks/add", params, headers)
        except socket.error as err:
            raise GPGUploadFailure(
                'Could not reach keyserver at http://%s %s' % (
                    keyserver_http_url, str(err)))

        assert conn.getresponse().status == httplib.OK, (
            'Keyserver POST failed')

        conn.close()

    def uploadPublicKey(self, fingerprint, logger=None):
        """Upload the specified public key to a keyserver.

        Use `retrieveKey` to get the public key content and upload an
        ASCII-armored export chunk.

        :param fingerprint: The key fingerprint, which must be an hexadecimal
            string.
        :raise GPGUploadFailure: if the keyserver could not be reaches.
        :raise AssertionError: if the POST request doesn't succeed.
        """
        if not config.getboolean('gpghandler', 'upload_keys'):
            if logger is not None:
                logger.info(
                    "Not submitting key to keyserver "
                    "(disabled in configuration).")
            return

        pub_key = self.retrieveKey(fingerprint)
        self._submitKey(pub_key.export())

    def getURLForKeyInServer(self, fingerprint, action='index', public=False):
        """Return the URL for that fingerprint on the configured keyserver.

        If public is True, return a URL for the public keyserver; otherwise,
        references the default (internal) keyserver.
        If action is provided, will attach that to the URL.
        """
        params = {
            'op': action,
            'search': '0x%s' % fingerprint,
            'fingerprint': 'on',
            }
        if public:
            host = config.get('gpghandler', 'public_host')
        else:
            host = config.get('gpghandler', 'host')
        return 'http://%s:%s/pks/lookup?%s' % (
            host, config.getint('gpghandler', 'port'),
            urllib.urlencode(sorted(params.items())))

    def _getPubKey(self, fingerprint):
        """Return the public key for the given fingerprint."""
        # request = get_current_browser_request()
        try:
            return self._grabPage('get', fingerprint)
        # We record an OOPS for most errors: If the keyserver does not
        # respond, callsites should show users an error message like
        # "sorry, the keyserver is not responding, try again in a few
        # minutes." The details of the error do not matter for users
        # (and for the code in callsites), but we should be able to see
        # if this problem occurs too often.
        except urllib2.HTTPError as exc:
            # Old versions of SKS return a 500 error when queried for a
            # non-existent key. Production was upgraded in 2013/01, but
            # let's leave this here for a while.
            #
            # We can extract the fact that the key is unknown by looking
            # into the response's content.
            if exc.code in (404, 500) and exc.fp is not None:
                content = exc.fp.read()
                no_key_message = 'No results found: No keys found'
                if content.find(no_key_message) >= 0:
                    raise GPGKeyDoesNotExistOnServer(fingerprint)
                # TODO: Find a replacement for errorlog
                # errorlog.globalErrorUtility.raising(sys.exc_info(), request)
                raise GPGKeyTemporarilyNotFoundError(fingerprint)
        except (socket.timeout, urllib2.URLError) as exc:
            # TODO: Find a replacement for errorlog
            # errorlog.globalErrorUtility.raising(sys.exc_info(), request)
            raise GPGKeyTemporarilyNotFoundError(fingerprint)

    def _grabPage(self, action, fingerprint):
        """Wrapper to collect KeyServer Pages."""
        url = self.getURLForKeyInServer(fingerprint, action)
        return urllib2.urlopen(
            url, timeout=config.getfloat('gpghandler', 'timeout')).read()


handler_singleton = GPGHandler()


class PymeSignature(object):
    """pyME signature container."""

    def __init__(self, fingerprint=None, plain_data=None, timestamp=None):
        """Initialized a signature container.

        :ivar fingerprint: Signer Fingerprint.
        :ivar plain_data: Plain Signed Text.
        :ivar timestamp: The time at which the message was signed.

        """
        self.fingerprint = fingerprint
        self.plain_data = plain_data
        self.timestamp = timestamp


class PymeKey:
    """pyME key model."""

    fingerprint = None
    exists_in_local_keyring = False

    def __init__(self, fingerprint):
        """Inititalize a key container.

        :ivar fingerprint: Key Fingerprint.
        :ivar algorithm: Key Algorithm.
        :ivar revoked: Key Revoked.
        :ivar expired: Key Expired.
        :ivar secret: Whether the key is secret of not.
        :ivar keysize: Key Size.
        :ivar keyid: Pseudo Key ID, composed by last fingerprint 8 digits.
        :ivar uids: List of user IDs associated with this key.
        :ivar emails: List containing only well formed and non-revoked emails.
        :ivar displayname: Key displayname: <size><type>/<keyid>.
        :ivar owner_trust: The owner trust.
        :ivar can_encrypt: Whether the key can be used for encrypting.
        :ivar can_sign: Whether the key can be used for signing.
        :ivar can_certify: Whether the key can be used for certification.
        :ivar can_authenticate: Whether the key can be used for authentication.

        """
        if fingerprint:
            self._buildFromFingerprint(fingerprint)

    @classmethod
    def newFromGpgmeKey(cls, key):
        """Initialize a PymeKey from a gpgme_key_t instance."""
        self = cls(None)
        self._buildFromGpgmeKey(key)
        return self

    def _buildFromFingerprint(self, fingerprint):
        """Build key information from a fingerprint."""
        context = gpgme.Context()
        # retrive additional key information
        try:
            key = context.get_key(fingerprint, False)
        except gpgme.GpgmeError:
            key = None

        if key and valid_fingerprint(key.subkeys[0].fpr):
            self._buildFromGpgmeKey(key)

    def _buildFromGpgmeKey(self, key):
        self.exists_in_local_keyring = True
        subkey = key.subkeys[0]
        self.fingerprint = subkey.fpr
        self.revoked = subkey.revoked
        self.keysize = subkey.length

        self.algorithm = GPGKeyAlgorithm.items[subkey.pubkey_algo].title
        self.keyid = self.fingerprint[-8:]
        self.expired = key.expired
        self.secret = key.secret
        self.owner_trust = key.owner_trust
        self.can_encrypt = key.can_encrypt
        self.can_sign = key.can_sign
        self.can_certify = key.can_certify
        self.can_authenticate = key.can_authenticate

        self.uids = [PymeUserId(uid) for uid in key.uids]

        # Non-revoked valid email addresses associated with this key
        self.emails = [uid.email for uid in self.uids
                       if valid_email(uid.email) and not uid.revoked]

    @property
    def displayname(self):
        return '%s%s/%s' % (self.keysize, self.algorithm, self.keyid)

    def export(self):
        """Export the context key in ASCII-armored mode.

        Both public and secret keys are supported, although secret keys are
        exported by calling `gpg` process while public ones use the native
        gpgme API.

        :return: a string containing the exported key.
        """
        if self.secret:
            # XXX cprov 20081014: gpgme_op_export() only supports public keys.
            # See http://www.fifi.org/cgi-bin/info2www?(gpgme)Exporting+Keys
            p = subprocess.Popen(
                ['gpg', '--export-secret-keys', '-a', self.fingerprint],
                stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            return p.stdout.read()

        context = gpgme.Context()
        context.armor = True
        keydata = StringIO()
        context.export(self.fingerprint.encode('ascii'), keydata)

        return keydata.getvalue()


class PymeUserId:
    """pyME user ID"""

    def __init__(self, uid):
        """
        :ivar revoked: True if the user ID has been revoked.
        :ivar invalid: True if the user ID is invalid.
        :ivar validity: A measure of the validity of the user ID, based on
            owner trust values and signatures.
        :ivar uid: A string identifying this user ID.
        :ivar name: The name portion of this user ID.
        :ivar email: The email portion of this user ID.
        :ivar comment: The comment portion of this user ID.
        """
        self.revoked = uid.revoked
        self.invalid = uid.invalid
        self.validity = uid.validity
        self.uid = uid.uid
        self.name = uid.name
        self.email = uid.email
        self.comment = uid.comment
