# -*- coding: utf-8 -*-
# Copyright 2016 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from __future__ import absolute_import
import collections
import datetime
import json
import logging
import logging.handlers
import os
import socket
import threading

__all__ = [
    'configure_logging'
]

_logging_configured = []


class SafeEncoder(json.JSONEncoder):
    """A JSON encoder that never fails to encode a field."""
    def default(self, obj):
        # try to get nice formatting for known types
        if isinstance(obj, datetime.datetime):
            return obj.isoformat()

        # fallback to just get object's repr
        try:
            return repr(obj)
        except:
            # but if repr fails, just get its info at 'object' level
            return "<<Unrepresentable: {} >>".format(object.__repr__(obj))


def safe_serializer(message):
    dumped = json.dumps(message, cls=SafeEncoder)
    return dumped.encode('utf8')


def configure_logging(config, service, level=logging.INFO, **kwds):
    """Configure service logging.

    If present, config['logstash'] is used to setup logstash logging.
    The `service` and optional kwds are passed as logstash entries.

    The log path is set to <LOG_DIR>/<service>.log assuming LOG_DIR is
    set in the environment.
    """
    # avoid duplicate logging
    if _logging_configured:
        return
    _logging_configured.append(True)

    if 'hostname' not in kwds:
        kwds['hostname'] = socket.gethostname()
    kwds['service'] = service

    class _LogstashExtraLogger(ExtraLogger):
        prefix = 'gpgservice.'
        extra_args = kwds
    logging.setLoggerClass(_LogstashExtraLogger)

    root_logger = logging.getLogger()
    root_logger.setLevel(level)

    if 'LOG_DIR' in os.environ:
        log_file_path = os.path.join(os.environ['LOG_DIR'], service + '.log')
        if os.path.exists(os.path.dirname(log_file_path)):
            handler = logging.handlers.TimedRotatingFileHandler(
                log_file_path,
                when='D',
                interval=1
            )
        else:
            print("parent directory for log '{}' does not exist, using stderr "
                  "for app log.".format(log_file_path))
            handler = logging.StreamHandler()
    else:
        print("No log file path given, using stderr for app log.")
        handler = logging.StreamHandler()

    handler.setFormatter(
        logging.Formatter(
            '%(asctime)s  %(name)s %(levelname)s: %(message)s'
        )
    )
    root_logger.addHandler(handler)
    if 'logstash' in config.sections():
        import logstash
        handler = logstash.LogstashHandler(
            config.get('logstash', 'host'),
            config.getint('logstash', 'port'),
            version=config.getint('logstash', 'version')
        )
        # patch the serializer in the handler's formatter to
        # use the own safe serializer
        handler.formatter.serialize = safe_serializer
        root_logger.addHandler(handler)


class ExtraLogger(logging.Logger):

    """A logger that handles passing 'extra' arguments to all logging calls.

    Tired of having to write:

    logger.info("Some message", extra=extra)

    ... everywhere?

    Now you can install this class as the Logger class:

    >>> import logging
    >>> logging.setLoggerClass(ExtraLogger)

    Then, to use it, get your logger as usual:

    >>> logger = logging.getLogger(__name__)

    ...and set your extra arguments once only:

    >>> logger.set_extra_args(dict(foo='bar', baz='123'))

    And those arguments will be included in all normal log messages:

    >>> logger.info("Hello World") # message will contain extra args set above

    Alternatively extra arguments can be attached to a subclass:

    >>> class XLogger(ExtraLogger):
    >>>     extra_args = dict(foo='bar', baz='123)

    Extra arguments can be passed to individual logging calls, and will
    take priority over any set with the set_extra_args call. Also a
    prefix can be specified using a subclass that will be put in front
    of the extra argument names for individual calls when storing them:

    >>> class PFXLogger(ExtraLogger):
    >>>     prefix = 'pfx.'

    """

    extra_args = {}
    prefix = ''

    def __init__(self, *args, **kwds):
        super(ExtraLogger, self).__init__(*args, **kwds)
        extra_args = self.extra_args

        # per thread extras
        class _perthread_extra(threading.local):
            def __init__(self):
                self.args = extra_args.copy()
        self._extra = _perthread_extra()

    def add_extra_args(self, extra_args):
        """Attach 'extra' arguments you want to be passed to every log message
           on the current thread.

        :param extra_args: A mapping type that contains the extra arguments you
            want to add.
        :raises TypeError: if extra_args is not a mapping type.

        """
        if not isinstance(extra_args, collections.Mapping):
            raise TypeError("extra_args must be a mapping")
        self._extra.args.update(extra_args)

    def reset_extra_args(self):
        """Reset all 'extra' arguments to every log message attached
           on the current thread.
        """
        self._extra.args = self.extra_args.copy()

    def makeRecord(self, name, level, fn, lno, msg, args, exc_info,
                   func=None, user_extra=None):
        extra = self._extra.args.copy()
        if user_extra:
            for k, v in user_extra.items():
                extra[self.prefix + k] = v
        return super(ExtraLogger, self).makeRecord(
            name, level, fn, lno, msg, args, exc_info, func, extra)
