# -*- coding: utf-8 -*-
# Copyright 2016 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

from testtools import TestCase
from testtools.matchers import (
    Contains,
    HasLength,
    )

from gpgservice import database


class DatabaseTests(TestCase):

    def setUp(self):
        super(DatabaseTests, self).setUp()
        self.connection = database.engine.connect()
        self.transaction = self.connection.begin()
        self.addCleanup(self.transaction.rollback)
        self.addCleanup(self.connection.close)

    def test_key_exists_returns_false_on_empty_database(self):
        self.assertFalse(database.key_exists(self.connection, 'nonexistant'))

    def test_insert_key(self):
        insert_key(self.connection, fingerprint='DEADBEEF')
        self.assertTrue(database.key_exists(self.connection, 'DEADBEEF'))

    def test_cannot_insert_duplicate_keys(self):
        insert_key(self.connection, fingerprint='FINGERPRINT')
        self.assertRaises(
            database.DuplicateKeyInsertionError,
            insert_key, self.connection, fingerprint='FINGERPRINT')

    def test_get_keys_by_owner_returns_empty_list(self):
        keys = list(database.get_keys_for_owner(self.connection, 'ownerid'))
        self.assertThat(keys, HasLength(0))

    def test_get_keys_retrieves_my_key(self):
        insert_key(self.connection, owner='me', key_id='1024D/17B05A8F',
                   fingerprint='84D205F03E1E67096CB54E262BE83793AACCD97C',
                   size=1024, algorithm='D')
        keys = list(database.get_keys_for_owner(self.connection, 'me'))
        expected_row = ('me', '1024D/17B05A8F',
                        '84D205F03E1E67096CB54E262BE83793AACCD97C', 1024, 'D',
                        True, False)
        self.assertThat(keys, HasLength(1))
        self.assertThat(keys, Contains(expected_row))
        key = keys[0]
        self.assertEqual('me', key.owner)
        self.assertEqual('1024D/17B05A8F', key.keyid)
        self.assertEqual('84D205F03E1E67096CB54E262BE83793AACCD97C',
                         key.fingerprint)
        self.assertEqual(1024, key.keysize)
        self.assertEqual('D', key.algorithm)
        self.assertEqual(True, key.enabled)
        self.assertEqual(False, key.can_encrypt)

    def test_get_keys_only_retrieves_my_keys(self):
        insert_key(self.connection, owner='me', key_id='1024D/17B05A8F',
                   fingerprint='84D205F03E1E67096CB54E262BE83793AACCD97C',
                   size=1024, algorithm='D')
        insert_key(self.connection, owner='other', key_id='1024D/DEADBEEF',
                   fingerprint='DEAD05F03E1E67096CB54E262BE83793AACCBEEF',
                   size=1024, algorithm='D')
        keys = list(database.get_keys_for_owner(self.connection, 'me'))
        expected_row = ('me', '1024D/17B05A8F',
                        '84D205F03E1E67096CB54E262BE83793AACCD97C', 1024, 'D',
                        True, False)
        self.assertThat(keys, HasLength(1))
        self.assertThat(keys, Contains(expected_row))

    def test_get_keys_can_retrieve_multiple_keys(self):
        insert_key(self.connection, owner='me', key_id='1024D/17B05A8F',
                   fingerprint='84D205F03E1E67096CB54E262BE83793AACCD97C',
                   size=1024, algorithm='D')
        insert_key(self.connection, owner='me', key_id='1024D/DEADBEEF',
                   fingerprint='DEAD05F03E1E67096CB54E262BE83793AACCBEEF',
                   size=1024, algorithm='D')
        keys = list(database.get_keys_for_owner(self.connection, 'me'))
        expected_key_one = ('me', '1024D/17B05A8F',
                            '84D205F03E1E67096CB54E262BE83793AACCD97C', 1024,
                            'D', True, False)
        expected_key_two = ('me', '1024D/DEADBEEF',
                            'DEAD05F03E1E67096CB54E262BE83793AACCBEEF', 1024,
                            'D', True, False)
        self.assertThat(keys, HasLength(2))
        self.assertThat(keys, Contains(expected_key_one))
        self.assertThat(keys, Contains(expected_key_two))

    def test_owner_owns_fingerprint_returns_false_on_unregistered_key(self):
        self.assertFalse(
            database.owner_owns_fingerprint(self.connection, 'owner',
                                            'fingerprint'))

    def test_owner_owns_fingerprint_returns_false_on_others_key(self):
        insert_key(self.connection, owner='other', fingerprint='fingerprint')
        self.assertFalse(
            database.owner_owns_fingerprint(self.connection, 'owner',
                                            'fingerprint'))

    def test_owner_owns_fingerprint_returns_true_on_our_key(self):
        insert_key(self.connection, owner='owner', fingerprint='fingerprint')
        self.assertTrue(
            database.owner_owns_fingerprint(self.connection, 'owner',
                                            'fingerprint'))

    def test_disable_key_works(self):
        insert_key(self.connection, owner='owner', fingerprint='fingerprint')
        self.assertTrue(
            database.disable_key(self.connection, 'fingerprint'))
        keys = list(database.get_keys_for_owner(self.connection, 'owner'))
        self.assertThat(keys, HasLength(1))
        self.assertEqual(False, keys[0].enabled)

    def test_disable_key_returns_false_on_bad_fingerprint(self):
        self.assertFalse(
            database.disable_key(self.connection, 'doesnotexist'))

    def test_enable_key_works(self):
        insert_key(self.connection, owner='owner',
                   fingerprint='fingerprint', enabled=False)
        self.assertTrue(
            database.enable_key(self.connection, 'fingerprint'))
        keys = list(database.get_keys_for_owner(self.connection, 'owner'))
        self.assertThat(keys, HasLength(1))
        self.assertEqual(True, keys[0].enabled)

    def test_enable_key_returns_false_on_bad_fingerprint(self):
        self.assertFalse(
            database.enable_key(self.connection, 'doesnotexist'))

    def test_get_key_by_fingerprint_returns_None_on_nonexistant_key(self):
        self.assertEqual(
            [],
            database.get_keys_by_fingerprints(self.connection,
                                              'doesnotexist'))

    def test_get_key_by_fingerprint_returns_key(self):
        insert_key(self.connection, fingerprint='fingerprint')
        keys = database.get_keys_by_fingerprints(self.connection,
                                                 'fingerprint')
        self.assertEqual('fingerprint', keys[0].fingerprint)


def insert_key(connection, owner="owner-identifier", key_id="key-identifier",
               fingerprint="DEADBEEF", size=1024, algorithm='D', enabled=True,
               can_encrypt=False):
    key = database.GPGKey(owner=owner, keyid=key_id, fingerprint=fingerprint,
                          keysize=size, algorithm=algorithm, enabled=enabled,
                          can_encrypt=can_encrypt)
    return database.insert_key(connection, key)


def test_suite():
    from unittest import TestLoader
    return TestLoader().loadTestsFromName(__name__)
