# -*- coding: utf-8 -*-
# Copyright 2009-2016 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).
import logging
import os
import os.path
import StringIO
from textwrap import dedent

import fixtures
import gpgme
from testtools import TestCase
from testtools.matchers import (
    DirContains,
    DirExists,
    EndsWith,
    HasLength,
    IsInstance,
    Not,
    StartsWith,
)

from gpgservice.handler import (
    GPGHandler,
    GPGKeyAlgorithm,
    handler_singleton,
    PymeKey,
)
from gpgservice.errors import (
    GPGKeyCannotEncrypt,
    GPGKeyDoesNotExistOnServer,
    GPGKeyExpired,
    GPGKeyNotFoundError,
    GPGKeyRevoked,
    GPGKeyTemporarilyNotFoundError,
    GPGUploadFailure,
    GPGVerificationError,
    MoreThanOneGPGKeyFound,
    SecretGPGKeyImportDetected,
    )
from gpgservice.config import set_config_value
from gpgservice.testing.keyserver import KeyServerTac
from gpgservice.tests.gpgkeys import (
    decrypt_content,
    import_secret_test_key,
    iter_test_key_emails,
    test_keyrings,
    test_pubkey_from_email,
    test_seckey_from_email,
    )


class BufferLogger(logging.Logger):

    def __init__(self):
        super(BufferLogger, self).__init__(__name__)
        self._handler = logging.StreamHandler(StringIO.StringIO())
        self.addHandler(self._handler)
        self._handler.setFormatter(
            logging.Formatter("%(levelname)s %(message)s")
        )

    def getLogBuffer(self):
        return self._handler.stream.getvalue()


class HandlerInfrastructureTests(TestCase):

    def test_home_dir_creation_and_destruction(self):
        self.useFixture(fixtures.EnvironmentVariable('GNUPGHOME'))
        handler = GPGHandler()

        self.assertThat(handler.home, DirExists())
        self.assertThat(handler.home, DirContains(['gpg.conf']))
        self.assertEqual(handler.home, os.environ.get('GNUPGHOME'))

        handler.removeHome()
        self.assertThat(handler.home, Not(DirExists()))


class GPGHandlerTestBase(TestCase):

    def setUp(self):
        super(GPGHandlerTestBase, self).setUp()
        self.gpg_handler = handler_singleton
        self.gpg_handler.resetLocalState()
        self.addCleanup(self.gpg_handler.resetLocalState)

    def populateKeyring(self):
        for email in iter_test_key_emails():
            pubkey = test_pubkey_from_email(email)
            self.gpg_handler.importPublicKey(pubkey)


class TestLocalKeyFiltering(GPGHandlerTestBase):

    """Tests for various filtering methods on localKeys()"""

    def test_localKeys_starts_empty(self):
        self.assertThat(list(self.gpg_handler.localKeys()), HasLength(0))

    def test_can_populate_localKeys(self):
        """Import our test keys and check they get imported."""
        self.populateKeyring()

        self.assertNotEqual([], list(self.gpg_handler.localKeys()))
        fingerprints = set(
            key.fingerprint for key in self.gpg_handler.localKeys()
        )
        self.assertIn(
            "340CA3BB270E2716C9EE0B768E7EB7086C64A8C5",
            fingerprints,
        )
        self.assertIn(
            "A419AE861E88BC9E04B9C26FBA2B9389DFD20543",
            fingerprints,
        )

    def test_local_key_filter_by_fingerprint(self):
        self.populateKeyring()
        target_fingerprint = '340CA3BB270E2716C9EE0B768E7EB7086C64A8C5'

        filtered_keys = self.gpg_handler.localKeys(filter=target_fingerprint)
        [key] = filtered_keys
        self.assertEqual(key.fingerprint, target_fingerprint)

    def test_local_key_filter_by_id(self):
        self.populateKeyring()
        target_fingerprint = '340CA3BB270E2716C9EE0B768E7EB7086C64A8C5'

        # Finding a key by its key ID.
        filtered_keys = self.gpg_handler.localKeys(target_fingerprint[-8:])
        [key] = filtered_keys
        self.assertEqual(key.fingerprint, target_fingerprint)

    def test_local_key_filter_by_email(self):
        self.populateKeyring()
        target_fingerprint = '340CA3BB270E2716C9EE0B768E7EB7086C64A8C5'

        # Multiple results when filtering by email.
        filtered_keys = self.gpg_handler.localKeys('foo.bar@canonical.com')

        filtered_fingerprints = [key.fingerprint for key in filtered_keys]
        self.assertIn(target_fingerprint, filtered_fingerprints)
        self.assertIn(
            'FD311613D941C6DE55737D310E3498675D147547',
            filtered_fingerprints,
        )

    def test_local_key_filter_by_secret_keys(self):
        self.assertEqual(list(self.gpg_handler.localKeys(secret=True)), [])

        import_secret_test_key()
        target_secret_fingerprint = 'A419AE861E88BC9E04B9C26FBA2B9389DFD20543'

        filtered_keys = self.gpg_handler.localKeys(secret=True)
        [key] = filtered_keys
        self.assertEqual(key.fingerprint, target_secret_fingerprint)

    def test_local_key_filter_by_secret_and_id(self):
        import_secret_test_key()
        target_secret_fingerprint = 'A419AE861E88BC9E04B9C26FBA2B9389DFD20543'

        filtered_keys = self.gpg_handler.localKeys(
            filter=target_secret_fingerprint[-8:], secret=True)

        [key] = filtered_keys
        self.assertEqual(key.fingerprint, target_secret_fingerprint)

    def test_local_key_filter_with_unicode_filter(self):
        self.populateKeyring()
        target_fingerprint = u'340CA3BB270E2716C9EE0B768E7EB7086C64A8C5'

        filtered_keys = self.gpg_handler.localKeys(target_fingerprint)
        [key] = filtered_keys
        self.assertEqual(key.fingerprint, target_fingerprint)

    def test_local_key_filter_with_non_ascii_filter(self):
        filtered_keys = self.gpg_handler.localKeys(u'non-ascii \u8463')
        self.failUnlessRaises(StopIteration, filtered_keys.next)

    def test_have_one_test_keyring(self):
        self.assertThat(list(test_keyrings()), HasLength(1))


class RetrieveKeyTests(GPGHandlerTestBase):

    def test_retrieveKey_raises_GPGKeyDoesNotExistOnServer(self):
        self.useFixture(KeyServerTac())
        self.assertRaises(
            GPGKeyDoesNotExistOnServer,
            self.gpg_handler.retrieveKey,
            'non-existent-fp',
        )

    def test_retrieveKey_raises__on_timeout(self):
        # If the keyserver responds too slowly, GPGHandler.retrieveKey()
        # raises GPGKeyTemporarilyNotFoundError.
        self.useFixture(KeyServerTac())
        with set_config_value('gpghandler', 'timeout', 0.01):
            self.assertRaises(
                GPGKeyTemporarilyNotFoundError,
                self.gpg_handler.retrieveKey,
                'non-existent-fp',
            )
            # An OOPS report is generated for the timeout.
            # TODO: Find a replacement for oops inspection here
            # error_report = self.oopses[-1]
            # self.assertEqual('TimeoutError', error_report['type'])
            # self.assertEqual('timeout exceeded.', error_report['value'])


class UploadKeyTests(GPGHandlerTestBase):

    def test_uploadPublicKey_suppress_in_config(self):
        self.useFixture(KeyServerTac())
        logger = BufferLogger()
        self.populateKeyring()
        fingerprint = list(self.gpg_handler.localKeys())[0].fingerprint

        with set_config_value('gpghandler', 'upload_keys', False):
            self.gpg_handler.uploadPublicKey(fingerprint, logger=logger)

        self.assertEqual(
            "INFO Not submitting key to keyserver "
            "(disabled in configuration).\n",
            logger.getLogBuffer()
        )
        self.assertRaises(
            GPGKeyDoesNotExistOnServer,
            self.gpg_handler._getPubKey,
            fingerprint
        )


class ImportPublicKeyTests(GPGHandlerTestBase):

    def test_import_raises_on_empty_public_key(self):
        self.assertRaises(
            GPGKeyNotFoundError,
            self.gpg_handler.importPublicKey,
            ''
        )

    def test_import_raises_on_bogus_public_key(self):
        self.assertRaises(
            GPGKeyNotFoundError,
            self.gpg_handler.importPublicKey,
            'XXXXXXXX'
        )

    def test_can_import_public_key(self):
        key_data = test_pubkey_from_email('test@canonical.com')
        key = self.gpg_handler.importPublicKey(key_data)

        self.assertThat(key, IsInstance(PymeKey))
        self.assertEqual(
            'A419AE861E88BC9E04B9C26FBA2B9389DFD20543',
            key.fingerprint
        )
        self.assertFalse(key.secret)
        self.assertTrue(key.can_encrypt)
        self.assertTrue(key.can_sign)
        self.assertTrue(key.can_certify)
        self.assertFalse(key.can_authenticate)

    def test_imported_keys_can_be_exported(self):
        key_data = test_pubkey_from_email('test@canonical.com')
        key = self.gpg_handler.importPublicKey(key_data)
        exported_key = key.export()

        self.assertThat(
            exported_key,
            StartsWith('-----BEGIN PGP PUBLIC KEY BLOCK-----')
        )
        self.assertThat(
            exported_key,
            EndsWith('-----END PGP PUBLIC KEY BLOCK-----\n')
        )

    def test_import_key_raises_on_secret_key(self):
        key_data = test_seckey_from_email('test@canonical.com')
        self.assertRaises(
            SecretGPGKeyImportDetected,
            self.gpg_handler.importPublicKey,
            key_data
        )

    def test_import_key_raises_on_multiple_keys(self):
        key_data = '\n'.join((
            test_pubkey_from_email('test@canonical.com'),
            test_pubkey_from_email('foo.bar@canonical.com'),
        ))
        self.assertRaises(
            MoreThanOneGPGKeyFound,
            self.gpg_handler.importPublicKey,
            key_data
        )

    def test_import_key_raises_on_damaged_preamble(self):
        key_data = test_pubkey_from_email('test@canonical.com')[1:]
        self.assertRaises(
            GPGKeyNotFoundError,
            self.gpg_handler.importPublicKey,
            key_data
        )

    def test_can_import_incomplete_pubkey(self):
        key_data = test_pubkey_from_email('test@canonical.com')[:-300]
        key = self.gpg_handler.importPublicKey(key_data)

        self.assertThat(key, IsInstance(PymeKey))
        self.assertEqual(
            'A419AE861E88BC9E04B9C26FBA2B9389DFD20543',
            key.fingerprint,
        )

    def test_cant_import_badly_broken_public_key(self):
        key_data = test_pubkey_from_email('test@canonical.com')[:-500]
        self.assertRaises(
            GPGKeyNotFoundError,
            self.gpg_handler.importPublicKey,
            key_data
        )


class ImportSecretKeyTests(GPGHandlerTestBase):

    def test_can_import_secret_key(self):
        key_data = test_seckey_from_email('test@canonical.com')

        key = self.gpg_handler.importSecretKey(key_data)

        self.assertThat(key, IsInstance(PymeKey))
        self.assertEqual(
            'A419AE861E88BC9E04B9C26FBA2B9389DFD20543',
            key.fingerprint,
        )
        self.assertTrue(key.secret)
        self.assertTrue(key.can_encrypt)
        self.assertTrue(key.can_sign)
        self.assertTrue(key.can_certify)
        self.assertFalse(key.can_authenticate)

    def test_secret_keys_can_be_exported(self):
        key_data = test_seckey_from_email('test@canonical.com')
        key = self.gpg_handler.importSecretKey(key_data)

        exported_data = key.export()

        self.assertThat(
            exported_data,
            StartsWith('-----BEGIN PGP PRIVATE KEY BLOCK-----')
        )
        self.assertThat(
            exported_data,
            EndsWith('-----END PGP PRIVATE KEY BLOCK-----\n')
        )

    def test_raises_when_importing_multiple_secret_keys(self):
        key_data = test_seckey_from_email('test@canonical.com')
        key_data += test_seckey_from_email('testing@canonical.com')
        self.assertRaises(MoreThanOneGPGKeyFound,
                          self.gpg_handler.importSecretKey,
                          key_data)

    def test_raises_with_no_content(self):
        self.assertRaises(GPGKeyNotFoundError,
                          self.gpg_handler.importSecretKey, '')


class GenerateKeyTests(GPGHandlerTestBase):

    """
    GPGHandler support GPG key generation.

    We are not testing real key generation because it depends on machine
    entropy that we may not have in PQM or EC2 (to not mention that it
    take minutes even when we have). So, the sample key in disk was
    generated by running the code below.

    We intentionally add some non-ascii characters in order to check if key
    generation and presentation cope with them.

        ### new_key = gpghandler.generateKey(
        ###     u"Launchpad PPA for Celso \xe1\xe9\xed\xf3\xfa Providelo")
        ### filepath = os.path.join(gpgkeysdir, 'ppa-sample@canonical.com.sec')
        ### export_file = open(filepath, 'w')
        ### export_file.write(new_key.export())
        ### export_file.close()

    XXX: It would be nice to find a better workaround to allow us to run these
         tests as part of autolanding.
    """

    def test_can_import_generated_secret_keys(self):
        key_data = test_seckey_from_email('ppa-sample@canonical.com')

        key = self.gpg_handler.importSecretKey(key_data)
        [uid] = key.uids

        self.assertTrue(key.secret)
        self.assertEqual('R', key.algorithm)
        self.assertEqual(1024, key.keysize)
        self.assertTrue(key.can_sign)
        self.assertFalse(key.can_encrypt)
        self.assertTrue(key.can_certify)
        self.assertFalse(key.can_authenticate)
        self.assertEqual(u'Launchpad PPA for Celso áéíóú Providelo', uid.name)
        self.assertEqual(u'', uid.comment)
        self.assertEqual(u'', uid.email)

    def test_can_retrieve_public_half_of_generated_key(self):
        key_data = test_seckey_from_email('ppa-sample@canonical.com')

        key = self.gpg_handler.importSecretKey(key_data)
        pubkey = self.gpg_handler.retrieveKey(key.fingerprint)
        [uid] = pubkey.uids

        self.assertFalse(pubkey.secret)
        self.assertEqual('R', pubkey.algorithm)
        self.assertEqual(1024, pubkey.keysize)
        self.assertFalse(pubkey.can_encrypt)
        self.assertTrue(pubkey.can_sign)
        self.assertEqual(u'Launchpad PPA for Celso áéíóú Providelo', uid.name)
        self.assertEqual(u'', uid.comment)
        self.assertEqual(u'', uid.email)


class KeyserverURLTests(GPGHandlerTestBase):

    def test_default_url_action(self):
        fingerprint = "A419AE861E88BC9E04B9C26FBA2B9389DFD20543"
        url = self.gpg_handler.getURLForKeyInServer(fingerprint)
        self.assertEqual(
            url,
            'http://localhost:11371/pks/lookup?'
            'fingerprint=on&'
            'op=index&'
            'search=0xA419AE861E88BC9E04B9C26FBA2B9389DFD20543'
        )

    def test_get_url_action(self):
        fingerprint = "A419AE861E88BC9E04B9C26FBA2B9389DFD20543"
        url = self.gpg_handler.getURLForKeyInServer(fingerprint, action="get")
        self.assertEqual(
            url,
            'http://localhost:11371/pks/lookup?'
            'fingerprint=on&'
            'op=get&'
            'search=0xA419AE861E88BC9E04B9C26FBA2B9389DFD20543'
        )

    def test_get_public_keyserver_location(self):
        fingerprint = "A419AE861E88BC9E04B9C26FBA2B9389DFD20543"
        url = self.gpg_handler.getURLForKeyInServer(fingerprint, public=True)
        self.assertEqual(
            url,
            'http://keyserver.ubuntu.com:11371/pks/lookup?'
            'fingerprint=on&'
            'op=index&'
            'search=0xA419AE861E88BC9E04B9C26FBA2B9389DFD20543'
        )


class PublicKeyUploadTests(GPGHandlerTestBase):

    def test_can_upload_and_retrieve_key(self):
        self.useFixture(KeyServerTac())
        key_data = test_seckey_from_email('ppa-sample@canonical.com')
        secret_key = self.gpg_handler.importSecretKey(key_data)

        self.gpg_handler.uploadPublicKey(secret_key.fingerprint)
        # delete all local keys, and make sure they're gone:
        self.gpg_handler.resetLocalState()
        self.assertThat(list(self.gpg_handler.localKeys()), HasLength(0))

        retrieved_key = self.gpg_handler.retrieveKey(secret_key.fingerprint)
        self.assertEqual(retrieved_key.fingerprint, secret_key.fingerprint)

    def test_upload_bad_key_raises(self):
        self.useFixture(KeyServerTac())
        self.assertRaises(
            GPGKeyDoesNotExistOnServer,
            self.gpg_handler.uploadPublicKey,
            'F' * 40
        )

    def test_can_upload_same_key_twice(self):
        self.useFixture(KeyServerTac())
        key_data = test_seckey_from_email('ppa-sample@canonical.com')
        secret_key = self.gpg_handler.importSecretKey(key_data)

        self.gpg_handler.uploadPublicKey(secret_key.fingerprint)
        self.gpg_handler.uploadPublicKey(secret_key.fingerprint)

    def test_unreachable_keyserver_upload_raises(self):
        key_data = test_seckey_from_email('ppa-sample@canonical.com')
        secret_key = self.gpg_handler.importSecretKey(key_data)

        self.assertRaises(
            GPGUploadFailure,
            self.gpg_handler.uploadPublicKey,
            secret_key.fingerprint
        )


class FingerprintSanitizationTests(GPGHandlerTestBase):

    def test_returns_none_on_short_fingerprints(self):
        self.assertIsNone(self.gpg_handler.sanitizeFingerprint('XXXXXX'))
        self.assertIsNone(self.gpg_handler.sanitizeFingerprint('681B 6469'))
        self.assertIsNone(self.gpg_handler.sanitizeFingerprint('abnckjdiue'))

    def test_removes_spaces(self):
        fingerprint = 'C858 2652 1A6E F6A6 037B  B3F7 9FF2 583E 681B 6469'
        expected = 'C85826521A6EF6A6037BB3F79FF2583E681B6469'
        self.assertEqual(
            expected,
            self.gpg_handler.sanitizeFingerprint(fingerprint)
        )

    def test_converts_to_uppercase(self):
        fingerprint = 'c858 2652 1a6e f6a6 037b  b3f7 9ff2 583e 681b 6469'
        expected = 'C85826521A6EF6A6037BB3F79FF2583E681B6469'
        self.assertEqual(
            expected,
            self.gpg_handler.sanitizeFingerprint(fingerprint)
        )

    def test_handles_non_ascii_characters_after_fingerprint(self):
        fingerprint = (u'c858 2652 1a6e f6a6 037b  b3f7 9ff2 583e 681b 6469 '
                       u'\xe9\xe1\xed')
        expected = 'C85826521A6EF6A6037BB3F79FF2583E681B6469'
        self.assertEqual(
            expected,
            self.gpg_handler.sanitizeFingerprint(fingerprint)
        )

    def test_disallows_non_ascii_characters_within_fingerprint(self):
        fingerprint = (
            u'\xe9\xe1\xed'
            u' c858 2652 1a6e f6a6 037b  b3f7 9ff2 583e 681b 6469 '
            u'\xe9\xe1\xed'
        )
        self.assertIsNone(self.gpg_handler.sanitizeFingerprint(fingerprint))


class GPGKeyAlgorithmTests(TestCase):

    def test_rsa(self):
        rsa_by_num = GPGKeyAlgorithm.items[1]
        rsa_by_name = GPGKeyAlgorithm.items['R']

        self.assertEqual(rsa_by_num, rsa_by_name)
        self.assertEqual("R", rsa_by_num.name)
        self.assertEqual("R", rsa_by_num.title)

    def test_elgamal(self):
        elg_by_num = GPGKeyAlgorithm.items[16]
        elg_by_name = GPGKeyAlgorithm.items['LITTLE_G']

        self.assertEqual(elg_by_num, elg_by_name)
        self.assertEqual("LITTLE_G", elg_by_num.name)
        self.assertEqual("g", elg_by_num.title)

    def test_dsa(self):
        dsa_by_num = GPGKeyAlgorithm.items[17]
        dsa_by_name = GPGKeyAlgorithm.items['D']

        self.assertEqual(dsa_by_num, dsa_by_name)
        self.assertEqual("D", dsa_by_num.name)
        self.assertEqual("D", dsa_by_num.title)

    def test_elgamal_compromised(self):
        elg_by_num = GPGKeyAlgorithm.items[20]
        elg_by_name = GPGKeyAlgorithm.items['G']

        self.assertEqual(elg_by_num, elg_by_name)
        self.assertEqual("G", elg_by_num.name)
        self.assertEqual("G", elg_by_num.title)


class GPGSignatureVerificationTests(GPGHandlerTestBase):

    def import_public_key_by_email(self, email):
        pubkey = test_pubkey_from_email(email)
        self.gpg_handler.importPublicKey(pubkey)

    def test_verify_signature_with_clearsigned(self):
        self.import_public_key_by_email('test@canonical.com')
        content = dedent("""\
            -----BEGIN PGP SIGNED MESSAGE-----
            Hash: SHA1

            Test Message.
            -----BEGIN PGP SIGNATURE-----
            Version: GnuPG v1.2.5 (GNU/Linux)

            iD8DBQFC7ZYY2yWXVgK6XvYRAgcwAJ43g/8X6DguixRucoXN86No67/W2wCgjFDj
            jLeauuXDPTcnzDmDzCaQLXo=
            =ettP
            -----END PGP SIGNATURE-----""")
        master_sig = self.gpg_handler.verifySignature(content)
        self.assertEqual(
            'A419AE861E88BC9E04B9C26FBA2B9389DFD20543',
            master_sig.fingerprint)

    def test_verify_signature_with_detached_binary_signature(self):
        self.import_public_key_by_email('test@canonical.com')
        content = 'Test content for detached signature.\n'
        signature = (
            b'\x88F\x04\x00\x11\x02\x00\x06\x05\x02V\xa0\x0c\xbd\x00\n\t\x10'
            b'\xdb%\x97V\x02\xba^\xf6\x8f]\x00\x9fu\x06HI\xb5\x0b\xec\xa4$\x11'
            b'\x9f\xaa\xe6\xf7\xd3\xdf\xafV8\x97\x00\xa0\x96\x9b\xef\x03\xaeh'
            b'\xc1\xe3:\xf8W\xc8\xe5\x8c\xe2\xe4?\xb6\x90\x18')
        signature = self.gpg_handler.verifySignature(content, signature)
        self.assertEqual(
            'A419AE861E88BC9E04B9C26FBA2B9389DFD20543',
            signature.fingerprint)

    def test_verify_signature_with_clearsigned_subkey(self):
        self.import_public_key_by_email('test@canonical.com')
        content = dedent("""\
            -----BEGIN PGP SIGNED MESSAGE-----
            Hash: SHA1

            Test Message.
            -----BEGIN PGP SIGNATURE-----
            Version: GnuPG v1.2.5 (GNU/Linux)

            iD8DBQFC7ZVH2yWXVgK6XvYRAqWEAJsF9I1MK2tRPPcxjXR2QclSPiQEsgCgtwst
            dZ7wZYeW68bk6GuuadabsSY=
            =ioKn
            -----END PGP SIGNATURE-----""")
        subkey_sig = self.gpg_handler.verifySignature(content)
        self.assertEqual(
            'A419AE861E88BC9E04B9C26FBA2B9389DFD20543',
            subkey_sig.fingerprint)

    def test_verify_signature_returns_None_on_bad_content(self):
        content = dedent("""\
            -----BEGIN PGP SIGNED MESSAGE-----
            Hash: SHA1

            Test Message.
            -----BEGIN PGP SIGNATURE-----
            Version: GnuPG v1.2.5 (GNU/Linux)

            iD8DBQFC7ZVH2yWXVgK6XvYRAqWEAJsF9I1MK2tRPPcxjXR2QclSPiQEsgCgtwst
            dZ7wZYeW68bk6GuuadabsSY=
            =ioKn
            -----END PGP SIGNATURE-----""")
        self.assertIsNone(self.gpg_handler.verifySignature(content))

    def test_verify_signature_with_tampered_clearsigned_content(self):
        self.import_public_key_by_email('test@canonical.com')
        content = dedent("""\
        -----BEGIN PGP SIGNED MESSAGE-----
        Hash: SHA1

        Test Message
        -----BEGIN PGP SIGNATURE-----
        Version: GnuPG v1.2.5 (GNU/Linux)

        iD8DBQFC7ZYY2yWXVgK6XvYRAgcwAJ43g/8X6DguixRucoXN86No67/W2wCgjFDj
        jLeauuXDPTcnzDmDzCaQLXo=
        =ettP
        -----END PGP SIGNATURE-----""")
        self.assertRaises(
            GPGVerificationError,
            self.gpg_handler.getVerifiedSignature,
            content)

    def test_verify_signature_with_no_signature(self):
        content = dedent("""\
        -----BEGIN PGP SIGNED MESSAGE-----
        Hash: SHA1

        Test Message
        -----BEGIN PGP SIGNATURE-----
        -----END PGP SIGNATURE-----""")
        self.assertRaises(
            GPGVerificationError,
            self.gpg_handler.getVerifiedSignature,
            content)

    def test_verify_signature_with_multiple_signatures(self):
        self.import_public_key_by_email('test@canonical.com')

        content = dedent("""\
            -----BEGIN PGP SIGNED MESSAGE-----
            Hash: SHA1

            This is the first piece of content.
            -----BEGIN PGP SIGNATURE-----
            Version: GnuPG v1

            iEYEARECAAYFAlagD/sACgkQ2yWXVgK6XvbIGACePu3yv3KtviLef0SC0CQ5GUXF
            ngIAn1vc7uk2KQBhKZbJLF2DmCWDgi84
            =kj7w
            -----END PGP SIGNATURE-----
            -----BEGIN PGP SIGNATURE-----
            Version: GnuPG v1

            iEYEARECAAYFAlagD/sACgkQ2yWXVgK6XvbIGACePu3yv3KtviLef0SC0CQ5GUXF
            ngIAn1vc7uk2KQBhKZbJLF2DmCWDgi84
            =kj7w
            -----END PGP SIGNATURE-----""")
        e = self.assertRaises(
            GPGVerificationError,
            self.gpg_handler.getVerifiedSignature,
            content)
        self.assertEqual(
            "Single signature expected, found multiple signatures",
            e.message)

    def test_verify_signature_with_multiple_messages(self):
        self.import_public_key_by_email('test@canonical.com')
        self.import_public_key_by_email('sign.only@canonical.com')

        content = dedent("""\
            -----BEGIN PGP SIGNED MESSAGE-----
            Hash: SHA1

            This is the first piece of content.
            -----BEGIN PGP SIGNATURE-----
            Version: GnuPG v1

            iEYEARECAAYFAlagD/sACgkQ2yWXVgK6XvbIGACePu3yv3KtviLef0SC0CQ5GUXF
            ngIAn1vc7uk2KQBhKZbJLF2DmCWDgi84
            =kj7w
            -----END PGP SIGNATURE-----
            -----BEGIN PGP SIGNED MESSAGE-----
            Hash: SHA1

            This is the second piece of content.
            -----BEGIN PGP SIGNATURE-----
            Version: GnuPG v1

            iEYEARECAAYFAlagFHIACgkQfYiRNxewWo/uKwCeMyO3+reLaoevBrk+ljk3WL4U
            j1QAn2YIkNLYLgXXo4hTqc5XohqxBgoC
            =SFMM
            -----END PGP SIGNATURE-----""")
        self.assertRaises(
            GPGVerificationError,
            self.gpg_handler.getVerifiedSignature,
            content)

    def test_verify_signature_with_missing_key(self):
        content = dedent("""\
            -----BEGIN PGP SIGNED MESSAGE-----
            Hash: SHA1

            Text Message
            -----BEGIN PGP SIGNATURE-----
            Version: GnuPG v1.4.2.2 (GNU/Linux)

            iD8DBQFFo+jp4ZLAVDsbsusRAhcYAJ9OOo7+tAxK94xGDu5yIUQG1LEY+wCeJvxr
            bOpYlIQD8vo7f9Y6LGqJbCc=
            =ds3K
            -----END PGP SIGNATURE-----""")
        e = self.assertRaises(
            GPGVerificationError,
            self.gpg_handler.getVerifiedSignature,
            content)
        self.assertEqual(e.message, u'No public key')

    def test_resilient_signature_verifiction_returns_all_errors(self):
        content = dedent("""\
            -----BEGIN PGP SIGNED MESSAGE-----
            Hash: SHA1

            Text Message
            -----BEGIN PGP SIGNATURE-----
            Version: GnuPG v1.4.2.2 (GNU/Linux)

            iD8DBQFFo+jp4ZLAVDsbsusRAhcYAJ9OOo7+tAxK94xGDu5yIUQG1LEY+wCeJvxr
            bOpYlIQD8vo7f9Y6LGqJbCc=
            =ds3K
            -----END PGP SIGNATURE-----""")
        e = self.assertRaises(
            GPGVerificationError,
            self.gpg_handler.getVerifiedSignatureResilient,
            content)

        msg = repr((7, 9, u'No public key'))
        self.assertEqual(
            'Verification failed 3 times: %r ' % [msg, msg, msg],
            e.message)

    def test_resilient_signature_verifiction(self):
        self.import_public_key_by_email('test@canonical.com')
        content = dedent("""\
            -----BEGIN PGP SIGNED MESSAGE-----
            Hash: SHA1

            Test Message.
            -----BEGIN PGP SIGNATURE-----
            Version: GnuPG v1.2.5 (GNU/Linux)

            iD8DBQFC7ZYY2yWXVgK6XvYRAgcwAJ43g/8X6DguixRucoXN86No67/W2wCgjFDj
            jLeauuXDPTcnzDmDzCaQLXo=
            =ettP
            -----END PGP SIGNATURE-----""")
        key = self.gpg_handler.getVerifiedSignatureResilient(content)
        self.assertEqual(
            'A419AE861E88BC9E04B9C26FBA2B9389DFD20543',
            key.fingerprint)


class GPGEncryptionTests(GPGHandlerTestBase):

    def test_returns_None_with_unknown_fingerprint(self):
        content = 'Hello World'
        cipher = self.gpg_handler.encryptContent(content, 'FEEDFACEDEAFBEEF')
        self.assertIsNone(cipher)

    def test_raises_when_using_sign_only_key(self):
        key = self.gpg_handler.importSecretKey(
            test_seckey_from_email('sign.only@canonical.com'))
        content = 'hello world'
        e = self.assertRaises(GPGKeyCannotEncrypt,
                              self.gpg_handler.encryptContent, content,
                              key.fingerprint)
        self.assertEqual(
            "key %s can not be used for encryption" % key.fingerprint,
            e.message)

    def test_can_encrypt_and_decrypt(self):
        key = self.gpg_handler.importSecretKey(
            test_seckey_from_email('test@canonical.com'))
        content = u'\ufcber'
        cipher = self.gpg_handler.encryptContent(
            content.encode('utf-8'),
            key.fingerprint)

        plain = decrypt_content(cipher, 'test').decode('utf-8')

        self.assertThat(cipher, StartsWith(b'-----BEGIN PGP MESSAGE-----'))
        self.assertThat(cipher, EndsWith(b'-----END PGP MESSAGE-----\n'))
        self.assertEqual(content, plain)

    def test_can_encrypt_other_charsets(self):
        key = self.gpg_handler.importSecretKey(
            test_seckey_from_email('test@canonical.com'))
        content = u'a\xe7ucar'
        cipher = self.gpg_handler.encryptContent(
            content.encode('iso-8859-1'),
            key.fingerprint)
        plain = decrypt_content(cipher, 'test').decode('iso-8859-1')
        self.assertEqual(content, plain)

    def test_raises_when_encrypting_unicode(self):
        key = self.gpg_handler.importSecretKey(
            test_seckey_from_email('test@canonical.com'))
        content = u'a\xe7ucar'
        self.assertRaises(
            TypeError,
            self.gpg_handler.encryptContent,
            content,
            key.fingerprint)

    def test_raises_when_decrypting_unicode(self):
        key = self.gpg_handler.importSecretKey(
            test_seckey_from_email('test@canonical.com'))
        content = u'\ufcber'
        cipher = self.gpg_handler.encryptContent(
            content.encode('utf-8'),
            key.fingerprint)

        cipher = unicode(cipher)
        self.assertRaises(TypeError, decrypt_content, cipher, 'test')

    def test_cannot_decrypt_content_with_unknown_key(self):
        cipher = dedent("""\
            -----BEGIN PGP MESSAGE-----
            Version: GnuPG v1.2.5 (GNU/Linux)

            hQEOAxkWbKKY/feTEAP+NoZRhb+/2OOd4f0FGRe7XP1HMvsRZ7X0uk/EjCv1CVnn
            TUS3UjPTWOsO27ehN6SeS0jXoD7o6LBOQ2b/P/mXor+nUAm9cb6OpE/FfYKINldp
            Hle2x6Zvp+/2Lc24wQ3TPnJYi+kEMTgOF0HUYNS32bOcmArZe7A6q2FKjamSHBkD
            /iLF0GRbkp8v9/NqGQ6JQBMn8YLJi8mXRCs4VsiqEV/sOpYkil6d7tRFN6vE3vOZ
            NUDiwbycPeArbrHZIQzmnYtwb/RHqpSJe1s9yiOc/OlA+IcRL/7nnNQMQx0mkiwL
            sDYPkggh2n4Q2Ekz1RuLenu7pMmDsc8rEqAmI1LeqnMF0sDVAU5NhqLp5ObBea2b
            G/b2jrAvKOIS0HkPITSuM6nGt5hyBpgfW0qgCYAuqy41g3VwveInaazsNDlEgFAI
            5OEmM2uvEKnTTfbwOqL/pPfKykemKvghUnyQRyig0zyMWr5WfvFbM/UTF6bw8utE
            AffSyBndPoGzJx7g7FHrRsgw17bvgSlSqA5XmwHq7qe0zc+FWN3S9ieb3aQR6/w0
            yIgU3hOHG469l8nV/35QH9xQRkQ7VPwIUMXDQCvTuSIg9uYbz8+TwJTwV5mLK/Nl
            FRowszYIbgELNtN+QsSpmZYvMl93p5wxmrIgmf8AlXpUwY5X1L8q+8a94zYCUhQS
            RQ1x4ejjULKx5HfX4dufIohCllAfMVFOkO7ywD7qDCOAIxyoC8oPWsXtTYqSwl5P
            nu/xOm3FroyjpMCe691BiMXxY7MsefyLt0kwIbCtzgp6btsg+96xz/S9LSqjPugA
            1rTfWQYZj/bmWcNaR5DSsQZAYalEGK+TI9hDA2ECruhYzMb8Ykl2c9FcdV3sOu9R
            UB+Czv0mWAmb
            =LQK5
            -----END PGP MESSAGE-----""")
        plain = decrypt_content(cipher, 'test')
        self.assertIsNone(plain)


class GPGContentSigningTests(GPGHandlerTestBase):

    def test_raises_with_unicode_content(self):
        key = self.gpg_handler.importSecretKey(
            test_seckey_from_email('test@canonical.com'))
        self.assertRaises(TypeError, self.gpg_handler.signContent,
                          u'Some unicode content', key.fingerprint)

    def test_can_clear_sign_content_with_general_key(self):
        key = self.gpg_handler.importSecretKey(
            test_seckey_from_email('test@canonical.com'))
        content = self.getUniqueString().encode('utf8')
        signature = self.gpg_handler.signContent(content, key.fingerprint,
                                                 'test')
        self.assertThat(signature, StartsWith(dedent("""\
            -----BEGIN PGP SIGNED MESSAGE-----
            Hash: SHA1

            %s
            -----BEGIN PGP SIGNATURE-----
            Version: GnuPG v1
            """ % content)))
        self.assertThat(signature, EndsWith('-----END PGP SIGNATURE-----\n'))

    def test_can_normal_sign_content_with_general_key(self):
        key = self.gpg_handler.importSecretKey(
            test_seckey_from_email('test@canonical.com'))
        content = self.getUniqueString()
        signature = self.gpg_handler.signContent(content, key.fingerprint,
                                                 'test', gpgme.SIG_MODE_NORMAL)

        self.assertThat(signature, StartsWith(dedent("""\
             -----BEGIN PGP MESSAGE-----
             Version: GnuPG v1
             """)))
        self.assertThat(signature, EndsWith('-----END PGP MESSAGE-----\n'))

    def test_can_detach_sign_content_with_general_key(self):
        key = self.gpg_handler.importSecretKey(
            test_seckey_from_email('test@canonical.com'))
        content = self.getUniqueString()
        signature = self.gpg_handler.signContent(content, key.fingerprint,
                                                 'test', gpgme.SIG_MODE_DETACH)
        self.assertThat(signature, StartsWith(dedent("""\
            -----BEGIN PGP SIGNATURE-----
            Version: GnuPG v1
            """)))
        self.assertThat(signature, EndsWith('-----END PGP SIGNATURE-----\n'))

    def test_can_verify_clearsigned_content(self):
        key = self.gpg_handler.importSecretKey(
            test_seckey_from_email('test@canonical.com'))
        content = self.getUniqueString().encode('utf8')
        signature = self.gpg_handler.signContent(content, key.fingerprint,
                                                 'test')
        signature = self.gpg_handler.verifySignature(signature)
        self.assertEqual(key.fingerprint, signature.fingerprint)

    def test_can_verify_normal_content(self):
        key = self.gpg_handler.importSecretKey(
            test_seckey_from_email('test@canonical.com'))
        content = self.getUniqueString().encode('utf8')
        signature = self.gpg_handler.signContent(content, key.fingerprint,
                                                 'test', gpgme.SIG_MODE_NORMAL)
        signature = self.gpg_handler.verifySignature(signature)
        self.assertEqual(key.fingerprint, signature.fingerprint)

    def test_can_verify_detached_content(self):
        key = self.gpg_handler.importSecretKey(
            test_seckey_from_email('test@canonical.com'))
        content = self.getUniqueString().encode('utf8')
        signature = self.gpg_handler.signContent(content, key.fingerprint,
                                                 'test', gpgme.SIG_MODE_DETACH)
        signature = self.gpg_handler.verifySignature(content, signature)
        self.assertEqual(key.fingerprint, signature.fingerprint)


class RetrieveActiveKeyTests(GPGHandlerTestBase):

    def test_raises_on_unknown_fingerprint(self):
        self.assertRaises(GPGKeyNotFoundError,
                          self.gpg_handler.retrieveActiveKey,
                          'FEEDFACEDEAFBEEF')

    def test_can_get_active_key(self):
        key = self.gpg_handler.importPublicKey(
            test_pubkey_from_email('test@canonical.com'))
        k = self.gpg_handler.retrieveActiveKey(key.fingerprint)
        self.assertEqual(key.fingerprint, k.fingerprint)

    def test_raises_on_revoked_key(self):
        key = self.gpg_handler.importPublicKey(
            test_pubkey_from_email('revoked.key@canonical.com'))
        self.assertRaises(
            GPGKeyRevoked, self.gpg_handler.retrieveActiveKey, key.fingerprint)

    def test_raises_on_expired_key(self):
        key = self.gpg_handler.importPublicKey(
            test_pubkey_from_email('expired.key@canonical.com'))
        self.assertRaises(
            GPGKeyExpired, self.gpg_handler.retrieveActiveKey, key.fingerprint)


def test_suite():
    from unittest import TestLoader
    return TestLoader().loadTestsFromName(__name__)
