# -*- coding: utf-8 -*-
# Copyright 2016 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).
from base64 import b64encode
import json
import random
import string
from StringIO import StringIO
from textwrap import dedent

from fixtures import Fixture
from psycopg2.extensions import AsIs
from sqlalchemy import sql
from testtools import (
    skipIf,
    TestCase,
    )
from testtools.matchers import (
    Contains,
    ContainsDict,
    EndsWith,
    Equals,
    HasLength,
    Matcher,
    MatchesAll,
    Mismatch,
    StartsWith,
    )

from gpgservice import (
    config,
    database,
    handler,
)
from gpgservice.testing.keyserver import KeyServerTac
from gpgservice.tests.test_database import insert_key
from gpgservice.webapp import create_flask_application


class GPGAppFixture(Fixture):

    def setUp(self):
        super(GPGAppFixture, self).setUp()
        self.app = create_flask_application(supress_logging=True)
        self.app.testing = True
        self.client = self.app.test_client()
        self._setup_database()
        self.addCleanup(handler.handler_singleton.resetLocalState)

    def __init__(self, recreate_engine=False):
        """Enable the engine to be recreated in tests where required."""
        self._recreate_engine = recreate_engine
        super(GPGAppFixture, self).__init__()

    def _setup_database(self):
        database.metadata.create_all()
        self.addCleanup(database.metadata.drop_all)

        if self._recreate_engine:
            original_engine = database.engine
            database.engine = database.create_engine()
            self.addCleanup(setattr, database, 'engine', original_engine)

        self.db_connection = database.engine.connect()
        self.addCleanup(self.db_connection.close)

    def post_json(self, path, data):
        """Make it easier to POST a json document to a URL."""
        return self.client.post(path, content_type='application/json',
                                data=json.dumps(data))

    def post_to_verify(self, document, signature=None):
        data = {
            'document': (StringIO(document),
                         'document',
                         'application/octet-stream')
        }
        if signature is not None:
            data['signature'] = (StringIO(signature),
                                 'signature',
                                 'application/pgp-signature')
        return self.client.post('/verify', data=data)

    def post_to_encrypt(self, fingerprint, content=b''):
        return self.client.post('/encrypt',
                                headers={'X-GPG-Fingerprint': fingerprint},
                                data=content,
                                content_type='application/octet-stream')


def get_random_user_id_pair():
    """Return a tuple of (decoded, encoded) user ids.

    The decoded user id will look similar to::

        https://login.ubuntu.com/+id/4tLsDY8

    ...while the encoded version will be base64 encoded with the '-_'
    characters replacing the '+/' characters.
    """
    candidates = string.ascii_lowercase + string.digits
    openid_id = ''.join((random.choice(candidates) for i in range(6)))
    unencoded = 'https://login.ubuntu.com/+id/' + openid_id
    encoded = b64encode(unencoded, altchars='-_')
    return unencoded, encoded


def get_random_fingerprint():
    choices = '0123456789ABCDEF'
    return ''.join(random.choice(choices) for i in range(40))


class WebAppListKeyTests(TestCase):

    def test_root(self):
        fixture = self.useFixture(GPGAppFixture())
        resp = fixture.client.get('/')
        self.assertEqual(200, resp.status_code)
        self.assertEqual("gpgservice - Copyright 2016 Canonical", resp.data)

    def test_get_keys_for_user_with_no_keys(self):
        fixture = self.useFixture(GPGAppFixture())
        user, encoded_user = get_random_user_id_pair()
        resp = fixture.client.get('/users/%s/keys' % encoded_user)
        self.assertEqual(200, resp.status_code)
        self.assertEqual('application/json', resp.content_type)
        data = json.loads(resp.data)
        expected = dict(user=user, keys=[])
        self.assertEqual(expected, data)

    def test_invalid_userid_returns_400(self):
        fixture = self.useFixture(GPGAppFixture())
        resp = fixture.client.get('/users/ /keys')
        self.assertThat(resp, IsAPIResponse(400, "Error: Invalid user id."))

    def test_get_keys_for_user_with_one_key(self):
        fixture = self.useFixture(GPGAppFixture())
        user, encoded_user = get_random_user_id_pair()
        insert_key(fixture.db_connection, owner=user,
                   key_id='1024D/17B05A8F',
                   fingerprint='84D205F03E1E67096CB54E262BE83793AACCD97C',
                   size=1024, algorithm='D')
        resp = fixture.client.get('/users/%s/keys' % encoded_user)
        self.assertEqual(200, resp.status_code)
        self.assertEqual('application/json', resp.content_type)
        data = json.loads(resp.data)
        self.assertEqual(data['user'], user)
        expected_key_display = {
            'id': '1024D/17B05A8F',
            'fingerprint': '84D205F03E1E67096CB54E262BE83793AACCD97C',
            'enabled': True,
            'size': 1024,
            'algorithm': 'D',
            'owner': user,
            'can_encrypt': False,
        }
        self.assertThat(data['keys'], HasLength(1), repr(data))
        self.assertEqual(data['keys'][0], expected_key_display)


class WebAppInsertKeyTests(TestCase):

    def assertKeyAddedSuccessfully(self, response, expected_status=201,
                                   **kwargs):
        self.assertThat(response.content_type, Equals('application/json'))
        self.assertThat(response.status_code, Equals(expected_status))
        data = json.loads(response.data)
        for key in kwargs:
            self.assertThat(data[key], Equals(kwargs[key]))

    def test_can_insert_fixture(self):
        self.useFixture(KeyServerTac())
        fixture = self.useFixture(GPGAppFixture())
        data = dict(
            fingerprint='447D BF38 C4F9 C4ED 7522  46B7 7D88 9137 17B0 5A8F')
        user, encoded_user = get_random_user_id_pair()
        resp = fixture.post_json('/users/%s/keys' % encoded_user, data)
        self.assertKeyAddedSuccessfully(
            resp,
            fingerprint='447DBF38C4F9C4ED752246B77D88913717B05A8F'
            )

    def test_inserting_key_adds_it_to_retrieval_api(self):
        self.useFixture(KeyServerTac())
        fixture = self.useFixture(GPGAppFixture())
        post_data = dict(
            fingerprint='447DBF38C4F9C4ED752246B77D88913717B05A8F')
        user, encoded_user = get_random_user_id_pair()
        resp = fixture.post_json('/users/%s/keys' % encoded_user, post_data)
        self.assertKeyAddedSuccessfully(resp)

        resp = fixture.client.get('/users/%s/keys' % encoded_user)
        self.assertEqual(200, resp.status_code)
        self.assertEqual('application/json', resp.content_type)
        data = json.loads(resp.data)
        self.assertThat(data['keys'], HasLength(1))
        self.assertEqual(post_data['fingerprint'],
                         data['keys'][0]['fingerprint'])

    def test_inserting_without_json_document_returns_400(self):
        fixture = self.useFixture(GPGAppFixture())
        _, encoded_user = get_random_user_id_pair()
        resp = fixture.client.post('/users/%s/keys' % encoded_user)
        self.assertThat(
            resp,
            IsAPIResponse(
                400, "Error: Missing JSON document in request body."))

    def test_inserting_with_missing_fingerprint_returns_400(self):
        fixture = self.useFixture(GPGAppFixture())
        _, encoded_user = get_random_user_id_pair()
        resp = fixture.post_json('/users/%s/keys' % encoded_user,
                                 dict(foo='bar'))
        self.assertThat(
            resp,
            IsAPIResponse(
                400, "Error: JSON document missing 'fingerprint' key."))

    def test_inserting_with_bad_fingerprint_returns_400(self):
        fixture = self.useFixture(GPGAppFixture())
        _, encoded_user = get_random_user_id_pair()
        resp = fixture.post_json('/users/%s/keys' % encoded_user,
                                 dict(fingerprint='FEEDFACE'))
        self.assertThat(
            resp,
            IsAPIResponse(400, "Error: Invalid fingerprint 'FEEDFACE'."))

    def test_inserting_without_keyserver_returns_503(self):
        fixture = self.useFixture(GPGAppFixture())
        data = dict(fingerprint='447DBF38C4F9C4ED752246B77D88913717B05A8F')
        _, encoded_user = get_random_user_id_pair()
        resp = fixture.post_json('/users/%s/keys' % encoded_user, data)
        self.assertThat(
            resp, IsAPIResponse(503, "Error: Keyserver unavailable."))

    def test_inserting_missing_key_returns_404(self):
        self.useFixture(KeyServerTac())
        fixture = self.useFixture(GPGAppFixture())
        data = dict(fingerprint='447DBF38C4F9C4ED752246B77D88913717B05A8D')
        user, encoded_user = get_random_user_id_pair()
        resp = fixture.post_json('/users/%s/keys' % encoded_user, data)
        self.assertThat(
            resp,
            IsAPIResponse(
                404, "Error: Fingerprint not found on keyserver."))

    def test_inserting_with_invalid_userid_returns_400(self):
        fixture = self.useFixture(GPGAppFixture())
        data = dict(fingerprint='447DBF38C4F9C4ED752246B77D88913717B05A8F')
        resp = fixture.post_json('/users/ \t /keys', data)
        self.assertThat(
            resp, IsAPIResponse(400, "Error: Invalid user id."))

    def test_inserting_duplicate_key_returns_409(self):
        self.useFixture(KeyServerTac())
        fixture = self.useFixture(GPGAppFixture())
        data = dict(fingerprint='447DBF38C4F9C4ED752246B77D88913717B05A8F')
        _, encoded_user1 = get_random_user_id_pair()
        resp = fixture.post_json('/users/%s/keys' % encoded_user1, data)

        self.assertKeyAddedSuccessfully(resp)

        _, encoded_user2 = get_random_user_id_pair()
        resp = fixture.post_json('/users/%s/keys' % encoded_user2, data)
        self.assertThat(
            resp,
            IsAPIResponse(409, "Error: Fingerprint already in database."))

    def test_inserting_with_others_disabled_key_returns_409(self):
        fixture = self.useFixture(GPGAppFixture())
        fingerprint = '447DBF38C4F9C4ED752246B77D88913717B05A8F'
        user1, _ = get_random_user_id_pair()
        _, encoded_user2 = get_random_user_id_pair()
        insert_key(fixture.db_connection, owner=user1,
                   fingerprint=fingerprint, enabled=False)
        resp = fixture.post_json('/users/%s/keys' % encoded_user2,
                                 dict(fingerprint=fingerprint))
        self.assertThat(
            resp,
            IsAPIResponse(409, "Error: Fingerprint already in database."))

    def test_inserting_with_others_activated_key_returns_409(self):
        fixture = self.useFixture(GPGAppFixture())
        fingerprint = '447DBF38C4F9C4ED752246B77D88913717B05A8F'
        user1, _ = get_random_user_id_pair()
        _, encoded_user2 = get_random_user_id_pair()
        insert_key(fixture.db_connection, owner=user1,
                   fingerprint=fingerprint, enabled=True)
        resp = fixture.post_json('/users/%s/keys' % encoded_user2,
                                 dict(fingerprint=fingerprint))
        self.assertThat(
            resp,
            IsAPIResponse(409, "Error: Fingerprint already in database."))

    def test_inserting_already_enabled_key_returns_200(self):
        fixture = self.useFixture(GPGAppFixture())
        fingerprint = '447DBF38C4F9C4ED752246B77D88913717B05A8F'
        user, encoded_user = get_random_user_id_pair()
        insert_key(fixture.db_connection, owner=user,
                   fingerprint=fingerprint, enabled=True)
        resp = fixture.post_json('/users/%s/keys' % encoded_user,
                                 dict(fingerprint=fingerprint))
        self.assertKeyAddedSuccessfully(
            resp,
            expected_status=200,
            owner=user,
            fingerprint=fingerprint,
            )

    def test_inserting_owned_disabled_key_returns_201(self):
        fixture = self.useFixture(GPGAppFixture())
        fingerprint = '447DBF38C4F9C4ED752246B77D88913717B05A8F'
        user, encoded_user = get_random_user_id_pair()
        insert_key(fixture.db_connection, owner=user,
                   fingerprint=fingerprint, enabled=False)
        resp = fixture.post_json('/users/%s/keys' % encoded_user,
                                 dict(fingerprint=fingerprint))
        self.assertKeyAddedSuccessfully(
            resp,
            owner=user,
            fingerprint=fingerprint,
            )


class WebAppDeactivateKeyTests(TestCase):

    def test_invalid_username_returns_400(self):
        fixture = self.useFixture(GPGAppFixture())
        resp = fixture.client.delete('/users/ \t /keys/deadbeef')
        self.assertThat(resp,
                        IsAPIResponse(400, "Error: Invalid user id."))

    def test_invalid_fingerprint_returns_400(self):
        fixture = self.useFixture(GPGAppFixture())
        _, encoded_user = get_random_user_id_pair()
        resp = fixture.client.delete('/users/%s/keys/deadbeef' % encoded_user)
        self.assertThat(resp,
                        IsAPIResponse(400, "Error: Invalid fingerprint."))

    def test_disabling_missing_key_returns_404(self):
        fixture = self.useFixture(GPGAppFixture())
        fingerprint = '447DBF38C4F9C4ED752246B77D88913717B05A8F'
        _, encoded_user = get_random_user_id_pair()
        resp = fixture.client.delete(
            '/users/%s/keys/%s' % (encoded_user, fingerprint))
        self.assertThat(
            resp, IsAPIResponse(404, "Error: Fingerprint not found."))

    def test_disabling_others_key_returns_404(self):
        fixture = self.useFixture(GPGAppFixture())
        fingerprint = '447DBF38C4F9C4ED752246B77D88913717B05A8F'
        user1, _ = get_random_user_id_pair()
        _, encoded_user2 = get_random_user_id_pair()
        insert_key(fixture.db_connection, owner=user1,
                   key_id='1024D/17B05A8F', fingerprint=fingerprint,
                   size=1024, algorithm='D')
        resp = fixture.client.delete(
            '/users/%s/keys/%s' % (encoded_user2, fingerprint))
        self.assertThat(
            resp, IsAPIResponse(404, "Error: Fingerprint not found."))

    def test_disabling_key_works(self):
        fixture = self.useFixture(GPGAppFixture())
        fingerprint = '447DBF38C4F9C4ED752246B77D88913717B05A8F'
        user, encoded_user = get_random_user_id_pair()
        insert_key(fixture.db_connection, owner=user,
                   key_id='1024D/17B05A8F', fingerprint=fingerprint,
                   size=1024, algorithm='D')
        resp = fixture.client.delete(
            '/users/%s/keys/%s' % (encoded_user, fingerprint))
        self.assertThat(resp, IsAPIResponse(200, "Key disabled"))


class KeyRetrievalTests(TestCase):

    def test_invalid_fingerprint_returns_400(self):
        fixture = self.useFixture(GPGAppFixture())
        bad_fingerprint = '44DBF384F9C4D75226B77D8913DEADBEF'
        resp = fixture.client.get('/keys/%s' % bad_fingerprint)
        self.assertThat(
            resp, IsAPIResponse(400, "Error: Invalid fingerprint."))

    def test_get_missing_fingerprint_returns_404(self):
        fixture = self.useFixture(GPGAppFixture())
        missing_fingerprint = '447DBF38C4F9C4ED752246B77D889137DEADBEEF'
        resp = fixture.client.get('/keys/%s' % missing_fingerprint)
        self.assertThat(
            resp, IsAPIResponse(404, "Error: Fingerprint not found."))

    def test_can_get_key(self):
        fixture = self.useFixture(GPGAppFixture())
        fingerprint = '447DBF38C4F9C4ED752246B77D88913717B05A8F'
        user, _ = get_random_user_id_pair()
        insert_key(fixture.db_connection, owner=user,
                   key_id='1024D/17B05A8F', fingerprint=fingerprint,
                   size=1024, algorithm='D')
        resp = fixture.client.get('/keys/%s' % fingerprint)
        self.assertThat(resp.status_code, Equals(200))
        self.assertThat(resp.content_type, Equals("application/json"))
        data = json.loads(resp.data)
        self.assertThat(
            data,
            Equals({'fingerprint': fingerprint,
                    'owner': user,
                    'algorithm': 'D',
                    'id': '1024D/17B05A8F',
                    'enabled': True,
                    'can_encrypt': False,
                    'size': 1024}))

    def test_can_get_multiple_keys(self):
        fixture = self.useFixture(GPGAppFixture())
        fingerprints = [get_random_fingerprint() for i in range(5)]
        user, _ = get_random_user_id_pair()
        for fingerprint in fingerprints:
            insert_key(fixture.db_connection, owner=user,
                       key_id='1024D/17B05A8F', fingerprint=fingerprint,
                       size=1024, algorithm='D')
        resp = fixture.client.get('/keys/%s' % ','.join(fingerprints))
        self.assertThat(resp.status_code, Equals(200))
        self.assertThat(resp.content_type, Equals('application/json'))
        data = json.loads(resp.data)
        returned_fingerprints = [k['fingerprint'] for k in data['keys']]
        for fingerprint in fingerprints:
            self.assertThat(returned_fingerprints, Contains(fingerprint))


class KeyServerRetrievalTests(TestCase):

    def test_returns_400_on_invalid_key(self):
        fixture = self.useFixture(GPGAppFixture())
        resp = fixture.client.get('/keyserver/BADBEEF')

        self.assertThat(
            resp, IsAPIResponse(400, "Error: Invalid fingerprint."))

    def test_returns_404_on_missing_key(self):
        self.useFixture(KeyServerTac())
        fixture = self.useFixture(GPGAppFixture())
        fingerprint = get_random_fingerprint()
        resp = fixture.client.get('/keyserver/%s' % fingerprint)

        self.assertThat(
            resp,
            IsAPIResponse(
                404, "Error: Key %r not found on keyserver." % fingerprint))

    def test_can_retrieve_regular_keyserver_key(self):
        self.useFixture(KeyServerTac())
        fixture = self.useFixture(GPGAppFixture())
        regular_key_fingerprint = 'A419AE861E88BC9E04B9C26FBA2B9389DFD20543'
        resp = fixture.client.get('/keyserver/%s' % regular_key_fingerprint)

        self.assertThat(resp.status_code, Equals(200))
        self.assertThat(resp.content_type, Equals('application/json'))
        data = json.loads(resp.data)
        self.assertThat(data, ContainsDict({
            'algorithm': Equals('D'),
            'can_authenticate': Equals(False),
            'can_certify': Equals(True),
            'can_encrypt': Equals(True),
            'can_sign': Equals(True),
            'displayname': Equals('1024D/DFD20543'),
            'emails': MatchesAll(
                Contains('sample.person@canonical.com'),
                Contains('test@canonical.com')),
            'expired': Equals(False),
            'fingerprint': Equals('A419AE861E88BC9E04B9C26FBA2B9389DFD20543'),
            'keyid': Equals('DFD20543'),
            'keysize': Equals(1024),
            'owner_trust': Equals(0),
            'revoked': Equals(False),
            'secret': Equals(False)
        }))

    def test_can_retrieve_sign_only_keyserver_key(self):
        self.useFixture(KeyServerTac())
        fixture = self.useFixture(GPGAppFixture())
        sign_only_key_fingerprint = '447DBF38C4F9C4ED752246B77D88913717B05A8F'
        resp = fixture.client.get('/keyserver/%s' % sign_only_key_fingerprint)

        self.assertThat(resp.status_code, Equals(200))
        self.assertThat(resp.content_type, Equals('application/json'))
        data = json.loads(resp.data)
        self.assertThat(data, ContainsDict({
            'algorithm': Equals('D'),
            'can_authenticate': Equals(False),
            'can_certify': Equals(True),
            'can_encrypt': Equals(False),
            'can_sign': Equals(True),
            'displayname': Equals('1024D/17B05A8F'),
            'emails': Equals(['sign.only@canonical.com']),
            'expired': Equals(False),
            'fingerprint': Equals('447DBF38C4F9C4ED752246B77D88913717B05A8F'),
            'keyid': Equals('17B05A8F'),
            'keysize': Equals(1024),
            'owner_trust': Equals(0),
            'revoked': Equals(False),
            'secret': Equals(False)
        }))

    def test_can_retrieve_revoked_keyserver_key(self):
        self.useFixture(KeyServerTac())
        fixture = self.useFixture(GPGAppFixture())
        revoked_key_fingerprint = '84D205F03E1E67096CB54E262BE83793AACCD97C'
        resp = fixture.client.get('/keyserver/%s' % revoked_key_fingerprint)

        self.assertThat(resp.status_code, Equals(200))
        self.assertThat(resp.content_type, Equals('application/json'))
        data = json.loads(resp.data)
        self.assertThat(data, ContainsDict({
            'algorithm': Equals('D'),
            'can_authenticate': Equals(False),
            'can_certify': Equals(True),
            'can_encrypt': Equals(False),
            'can_sign': Equals(True),
            'displayname': Equals('1024D/AACCD97C'),
            'emails': Equals([]),
            'expired': Equals(False),
            'fingerprint': Equals('84D205F03E1E67096CB54E262BE83793AACCD97C'),
            'keyid': Equals('AACCD97C'),
            'keysize': Equals(1024),
            'owner_trust': Equals(0),
            'revoked': Equals(True),
            'secret': Equals(False)
        }))

    def test_can_retrieve_expired_keyserver_key(self):
        self.useFixture(KeyServerTac())
        fixture = self.useFixture(GPGAppFixture())
        expired_key_fingerprint = 'ECA5B797586F2E27381A16CFDE6C9167046C6D63'
        resp = fixture.client.get('/keyserver/%s' % expired_key_fingerprint)

        self.assertThat(resp.status_code, Equals(200))
        self.assertThat(resp.content_type, Equals('application/json'))
        data = json.loads(resp.data)
        self.assertThat(data, ContainsDict({
            'algorithm': Equals('D'),
            'can_authenticate': Equals(False),
            'can_certify': Equals(True),
            'can_encrypt': Equals(False),
            'can_sign': Equals(True),
            'displayname': Equals('1024D/046C6D63'),
            'emails': Equals(['expired.key@canonical.com']),
            'expired': Equals(True),
            'fingerprint': Equals('ECA5B797586F2E27381A16CFDE6C9167046C6D63'),
            'keyid': Equals('046C6D63'),
            'keysize': Equals(1024),
            'owner_trust': Equals(0),
            'revoked': Equals(False),
            'secret': Equals(False)
        }))


class WebAppHeaderTests(TestCase):

    def test_includes_git_revision(self):
        fixture = self.useFixture(GPGAppFixture())
        resp = fixture.client.get('/')
        self.assertThat(resp.headers, Contains('X-GPGService-Revision'))

    def test_generates_new_request_id(self):
        fixture = self.useFixture(GPGAppFixture())
        resp = fixture.client.get('/')
        self.assertThat(resp.headers, Contains('X-Request-Id'))

    def test_preserves_existing_request_id(self):
        fixture = self.useFixture(GPGAppFixture())
        rid = self.getUniqueString()
        resp = fixture.client.get('/', headers={'X-Request-Id': rid})
        self.assertEqual(rid, resp.headers['X-Request-Id'])


class VerifyDocumentTests(TestCase):

    def test_returns_400_with_missing_document(self):
        fixture = self.useFixture(GPGAppFixture())
        resp = fixture.client.post('/verify')
        self.assertThat(
            resp,
            IsAPIResponse(400, "Error: Missing document to verify."))

    def test_returns_400_on_document_with_no_signature(self):
        fixture = self.useFixture(GPGAppFixture())
        resp = fixture.post_to_verify("unverified content")
        self.assertThat(resp, IsAPIResponse(
            400, "Error: Could not find any GPG verified content."), resp.data)

    def test_verifying_with_clearsigned(self):
        self.useFixture(KeyServerTac())
        fixture = self.useFixture(GPGAppFixture())
        document = dedent('''\
        -----BEGIN PGP SIGNED MESSAGE-----
        Hash: SHA1

        This is a test document.
        -----BEGIN PGP SIGNATURE-----
        Version: GnuPG v1

        iEYEARECAAYFAlauvhMACgkQfYiRNxewWo9RXgCeNTm2LHPTklDfkhPNDXlZZO8w
        +PgAn0jxaoVxm3cftv8VfHzs+2HfXU0H
        =3rvY
        -----END PGP SIGNATURE-----
        ''')
        resp = fixture.post_to_verify(document)
        self.assertEqual(200, resp.status_code)
        self.assertEqual("This is a test document.\n", resp.data)
        self.assertEqual("447DBF38C4F9C4ED752246B77D88913717B05A8F",
                         resp.headers['X-GPG-Verified-Fingerprint'])

    def test_verify_document_with_unverified_contents(self):
        self.useFixture(KeyServerTac())
        fixture = self.useFixture(GPGAppFixture())
        document = dedent('''\
        This part has not been signed.
        -----BEGIN PGP SIGNED MESSAGE-----
        Hash: SHA1

        This is a test document.
        -----BEGIN PGP SIGNATURE-----
        Version: GnuPG v1

        iEYEARECAAYFAlauvhMACgkQfYiRNxewWo9RXgCeNTm2LHPTklDfkhPNDXlZZO8w
        +PgAn0jxaoVxm3cftv8VfHzs+2HfXU0H
        =3rvY
        -----END PGP SIGNATURE-----
        Neither has this part!
        ''')
        resp = fixture.post_to_verify(document)
        self.assertEqual(200, resp.status_code)
        self.assertEqual("This is a test document.\n", resp.data)
        self.assertEqual("447DBF38C4F9C4ED752246B77D88913717B05A8F",
                         resp.headers['X-GPG-Verified-Fingerprint'])

    def test_verifying_document_with_detached_signature(self):
        self.useFixture(KeyServerTac())
        fixture = self.useFixture(GPGAppFixture())
        document = 'This is a test document.\n'
        signature = (b"\x88F\x04\x00\x11\x02\x00\x06\x05\x02V\xae\xbe'"
                     b"\x00\n\t\x10}\x88\x917\x17\xb0Z\x8f\xd5A\x00\x9f"
                     b"s#\xc1n\xb9\xc5h\xb0!\xb1\xb6\x97\x9d\xdb\xe8\x81"
                     b"\xf2\xd4\x88\xfc\x00\x9e6+\xb5\xfe\x10~V\x08!\x99,"
                     b"\x06\xe2\xe1\x86b\xf6\xa0\x92\xfe")
        resp = fixture.post_to_verify(document, signature)
        self.assertEqual(200, resp.status_code)
        self.assertEqual("This is a test document.\n", resp.data)
        self.assertEqual("447DBF38C4F9C4ED752246B77D88913717B05A8F",
                         resp.headers['X-GPG-Verified-Fingerprint'])

    def test_verify_document_with_unknown_key(self):
        self.useFixture(KeyServerTac())
        fixture = self.useFixture(GPGAppFixture())
        document = dedent('''\
        -----BEGIN PGP SIGNED MESSAGE-----
        Hash: SHA1

        This is a test document.
        -----BEGIN PGP SIGNATURE-----
        Version: GnuPG v1

        iQEcBAEBAgAGBQJWsUJVAAoJELsvu710zTSIC54IAJrsV3wOZQ62YQRdjSo4Q1Cl
        173W4HZOdfMKCgORI7K4N2SIMHM4EzBiGUHknZTQm9HV8qv0E0FOCDHkzQuVwKV4
        /elcCupBIecOZXKp/KLSJuctcxM5z5Xj49Hh9hYB2phhejkXJ31PHaW3nxeApSRf
        lh1ix9lhbnGzc+FkWVUfy0Sfcl9v5luE6R7c48CSV+APg7aImcVXiJWO6pwViLXm
        z84Sdj3cx5cMFkbC+KWogiHh8aOuuCEgWl+cEPysM/fMBu+PlJGF5c6pQ4sJUTqS
        WbZP8HyM2p2pQarf7EaMhAmsltLuirTdfmlmI4CUtQ5HSYqqs96hXVDL+70JQIs=
        =JXjx
        -----END PGP SIGNATURE-----''')
        resp = fixture.post_to_verify(document)
        self.assertThat(
            resp,
            IsAPIResponse(400,
                          "Error: Could not find GPG public key."))

    def test_verify_with_bad_clearsigned_document(self):
        self.useFixture(KeyServerTac())
        fixture = self.useFixture(GPGAppFixture())
        document = dedent('''\
        -----BEGIN PGP SIGNED MESSAGE-----
        Hash: SHA1

        This is a test document.
        -----BEGIN PGP SIGNATURE-----
        Version: GnuPG v1

        FEEDFACEAAYFAlauvhMACgkQfYiRNxewWo9RXgCeNTm2LHPTklDfkhPNDXlZZO8w
        +PgAn0jxaoVxm3cftv8VfHzsDEADBEEF
        =3rvY
        -----END PGP SIGNATURE-----
        ''')
        resp = fixture.post_to_verify(document)
        self.assertThat(
            resp,
            IsAPIResponse(400,
                          "Error: Could not find any GPG verified content."))

    def test_with_clearsigned_document_missing_signatures(self):
        self.useFixture(KeyServerTac())
        fixture = self.useFixture(GPGAppFixture())
        document = dedent('''\
        -----BEGIN PGP SIGNED MESSAGE-----
        Hash: SHA1

        This is a test document.
        -----BEGIN PGP SIGNATURE-----
        -----END PGP SIGNATURE-----
        ''')
        resp = fixture.post_to_verify(document)
        self.assertThat(
            resp,
            IsAPIResponse(400, "Error: No signatures found."))

    def test_with_clearsigned_document_multiple_signatures(self):
        self.useFixture(KeyServerTac())
        fixture = self.useFixture(GPGAppFixture())
        document = dedent('''\
        -----BEGIN PGP SIGNED MESSAGE-----
        Hash: SHA1

        This is a test document.
        -----BEGIN PGP SIGNATURE-----
        Version: GnuPG v1

        iEYEARECAAYFAlauvhMACgkQfYiRNxewWo9RXgCeNTm2LHPTklDfkhPNDXlZZO8w
        +PgAn0jxaoVxm3cftv8VfHzs+2HfXU0H
        =3rvY
        -----END PGP SIGNATURE-----
        -----BEGIN PGP SIGNATURE-----
        Version: GnuPG v1

        iEYEARECAAYFAlauvhMACgkQfYiRNxewWo9RXgCeNTm2LHPTklDfkhPNDXlZZO8w
        +PgAn0jxaoVxm3cftv8VfHzs+2HfXU0H
        =3rvY
        -----END PGP SIGNATURE-----
        ''')
        resp = fixture.post_to_verify(document)
        self.assertThat(
            resp,
            IsAPIResponse(400,
                          "Error: Single signature expected, "
                          "found multiple signatures."))

    def test_verify_with_large_document_fails(self):
        self.useFixture(KeyServerTac())
        document = dedent('''\
        -----BEGIN PGP SIGNED MESSAGE-----
        Hash: SHA1

        This is a test document.
        -----BEGIN PGP SIGNATURE-----
        Version: GnuPG v1

        iEYEARECAAYFAlauvhMACgkQfYiRNxewWo9RXgCeNTm2LHPTklDfkhPNDXlZZO8w
        +PgAn0jxaoVxm3cftv8VfHzs+2HfXU0H
        =3rvY
        -----END PGP SIGNATURE-----
        ''')
        with config.set_config_value('gpghandler', 'maximum_upload_size', 1):
            fixture = self.useFixture(GPGAppFixture())
            resp = fixture.post_to_verify(document)
            self.assertEqual(413, resp.status_code)


class EncryptDocumentTests(TestCase):

    def test_returns_400_if_key_not_found(self):
        fixture = self.useFixture(GPGAppFixture())
        resp = fixture.client.post('/encrypt')
        self.assertThat(
            resp,
            IsAPIResponse(400, 'Error: Missing X-GPG-Fingerprint header.'))

    def test_returns_400_on_bad_fingerprint(self):
        fixture = self.useFixture(GPGAppFixture())
        fingerprint = "invalid fingerprint"
        resp = fixture.post_to_encrypt(fingerprint)
        self.assertThat(
            resp,
            IsAPIResponse(
                400, "Error: Invalid fingerprint '%s'." % fingerprint))

    def test_returns_400_on_empty_content(self):
        fixture = self.useFixture(GPGAppFixture())
        fingerprint = 'A419 AE86 1E88 BC9E 04B9 C26F BA2B 9389 DFD2 0543'
        resp = fixture.post_to_encrypt(fingerprint)
        self.assertThat(resp,
                        IsAPIResponse(400, "Error: No content to encrypt."))

    def test_returns_400_on_unknown_key(self):
        fixture = self.useFixture(GPGAppFixture())
        self.useFixture(KeyServerTac())
        fingerprint = 'A419 AE86 1E88 BC9E 04B9 C26F BA2B 9389 DFD2 BEEF'
        resp = fixture.post_to_encrypt(fingerprint, b'Hello')
        self.assertThat(
            resp,
            IsAPIResponse(
                400, "Error: Could not retrieve key '%s'." % fingerprint))

    def test_can_encrypt_content(self):
        fixture = self.useFixture(GPGAppFixture())
        self.useFixture(KeyServerTac())
        fingerprint = 'A419 AE86 1E88 BC9E 04B9 C26F BA2B 9389 DFD2 0543'
        resp = fixture.post_to_encrypt(fingerprint, b'Hello')
        data = resp.data
        self.assertThat(data, StartsWith('-----BEGIN PGP MESSAGE-----'))
        self.assertThat(data, EndsWith('-----END PGP MESSAGE-----\n'))

    def test_returns_400_on_revoked_key(self):
        fixture = self.useFixture(GPGAppFixture())
        self.useFixture(KeyServerTac())
        fingerprint = '84D2 05F0 3E1E 6709 6CB5  4E26 2BE8 3793 AACC D97C'
        resp = fixture.post_to_encrypt(fingerprint, b'Hello')
        self.assertThat(
            resp,
            IsAPIResponse(
                400,
                "Error: GPG key '%s' has been revoked." % fingerprint))

    def test_returns_400_on_expired_key(self):
        fixture = self.useFixture(GPGAppFixture())
        self.useFixture(KeyServerTac())
        fingerprint = 'ECA5 B797 586F 2E27 381A  16CF DE6C 9167 046C 6D63'
        resp = fixture.post_to_encrypt(fingerprint, b'Hello')
        self.assertThat(
            resp,
            IsAPIResponse(
                400, "Error: GPG key '%s' has expired." % fingerprint))

    def test_returns_400_with_sign_only_key(self):
        fixture = self.useFixture(GPGAppFixture())
        self.useFixture(KeyServerTac())
        fingerprint = '447D BF38 C4F9 C4ED 7522  46B7 7D88 9137 17B0 5A8F'
        resp = fixture.post_to_encrypt(fingerprint, b'Hello')
        self.assertThat(
            resp,
            IsAPIResponse(
                400,
                "Error: GPG key '%s' cannot be used to encrypt." % fingerprint)
            )

    def test_returns_413_on_large_document(self):
        self.useFixture(KeyServerTac())
        fingerprint = '447D BF38 C4F9 C4ED 7522  46B7 7D88 9137 17B0 5A8F'
        with config.set_config_value('gpghandler', 'maximum_upload_size', 1):
            fixture = self.useFixture(GPGAppFixture())
            resp = fixture.post_to_encrypt(fingerprint, b'Hello')
            self.assertEqual(413, resp.status_code)


class WebAppTestApiDBResetTests(TestCase):

    def test_reset_db_disabled_from_config(self):
        with config.set_config_value('gpghandler', 'enable_test_endpoint',
                                     False):
            fixture = self.useFixture(GPGAppFixture())
            resp = fixture.client.post('/test/reset_db')
            self.assertEqual(404, resp.status_code)

    def test_reset_db_clears_all_data(self):
        fingerprint = '84D205F03E1E67096CB54E262BE83793AACCD97C'
        with config.set_config_value('gpghandler', 'enable_test_endpoint',
                                     True):
            fixture = self.useFixture(GPGAppFixture())
            user, _ = get_random_user_id_pair()
            insert_key(fixture.db_connection, owner=user,
                       key_id='1024D/17B05A8F', fingerprint=fingerprint,
                       size=1024, algorithm='D')
            resp = fixture.client.post('/test/reset_db')
            self.assertEqual(200, resp.status_code)
            self.assertFalse(
                database.key_exists(fixture.db_connection, fingerprint))

    def test_reset_db_can_insert_test_data(self):
        fingerprint = '84D205F03E1E67096CB54E262BE83793AACCD97C'
        with config.set_config_value('gpghandler', 'enable_test_endpoint',
                                     True):
            fixture = self.useFixture(GPGAppFixture())
            resp = fixture.post_json(
                '/test/reset_db',
                {'keys': [dict(owner='myrandomid',
                               id='1024D/17B05A8F',
                               fingerprint=fingerprint,
                               size=1024, algorithm='D')]})
            self.assertEqual(200, resp.status_code)
            self.assertTrue(
                database.key_exists(fixture.db_connection, fingerprint))

    def test_reset_db_can_insert_multiple_keys(self):
        key_one = dict(owner='myrandomid',
                       id='1024D/17B05A8F',
                       fingerprint='84D205F03E1E67096CB54E262BE83793AACCD97C',
                       size=1024, algorithm='D')
        key_two = dict(owner='foo bar',
                       id='1024D/17B06B90',
                       fingerprint='84D205F03E1E67096CB54E262BE8379317B06B90',
                       size=1024, algorithm='D')
        with config.set_config_value('gpghandler', 'enable_test_endpoint',
                                     True):
            fixture = self.useFixture(GPGAppFixture())
            resp = fixture.post_json(
                '/test/reset_db',
                {'keys': [key_one, key_two]})
            self.assertEqual(200, resp.status_code)
            self.assertTrue(
                database.key_exists(
                    fixture.db_connection, key_one['fingerprint']))
            self.assertTrue(
                database.key_exists(
                    fixture.db_connection, key_two['fingerprint']))

    def test_reset_db_returns_400_on_missing_key_data(self):
        fingerprint = '84D205F03E1E67096CB54E262BE83793AACCD97C'
        with config.set_config_value('gpghandler', 'enable_test_endpoint',
                                     True):
            fixture = self.useFixture(GPGAppFixture())
            user, _ = get_random_user_id_pair()
            insert_key(fixture.db_connection, owner=user,
                       key_id='1024D/17B05A8F', fingerprint=fingerprint,
                       size=1024, algorithm='D')
            resp = fixture.post_json('/test/reset_db', dict(foo='bar'))
            self.assertThat(
                resp,
                IsAPIResponse(
                    400,
                    "Error: JSON Content-Type set but missing 'keys' key."))
            self.assertTrue(
                database.key_exists(fixture.db_connection, fingerprint))

    def test_reset_db_returns_400_on_badly_formatted_key_data(self):
        fingerprint = '84D205F03E1E67096CB54E262BE83793AACCD97C'
        with config.set_config_value('gpghandler', 'enable_test_endpoint',
                                     True):
            fixture = self.useFixture(GPGAppFixture())
            user, _ = get_random_user_id_pair()
            insert_key(fixture.db_connection, owner=user,
                       key_id='1024D/17B05A8F', fingerprint=fingerprint,
                       size=1024, algorithm='D')
            resp = fixture.post_json(
                '/test/reset_db',
                {'keys': [dict(ownder='myrandomid',
                               ib='1024D/17B05A8F',
                               figgerprint=fingerprint,
                               siz=1024, gorithm='D')]})
            self.assertThat(resp, IsAPIResponse(400, "Error: Bad Key format."))
            self.assertTrue(
                database.key_exists(fixture.db_connection, fingerprint))


class WebAppTestApiDBInsertTests(TestCase):

    def test_db_insert_disabled_from_config(self):
        with config.set_config_value('gpghandler', 'enable_test_endpoint',
                                     False):
            fixture = self.useFixture(GPGAppFixture())
            resp = fixture.client.post('/test/add_keys')
            self.assertEqual(404, resp.status_code)

    def test_can_insert_test_data(self):
        fingerprint = '84D205F03E1E67096CB54E262BE83793AACCD97C'
        with config.set_config_value('gpghandler', 'enable_test_endpoint',
                                     True):
            fixture = self.useFixture(GPGAppFixture())
            resp = fixture.post_json(
                '/test/add_keys',
                {'keys': [dict(owner='myrandomid',
                               id='1024D/17B05A8F',
                               fingerprint=fingerprint,
                               size=1024, algorithm='D')]})
            self.assertEqual(200, resp.status_code)
            self.assertTrue(
                database.key_exists(fixture.db_connection, fingerprint))

    def test_can_insert_multiple_keys(self):
        key_one = dict(owner='myrandomid',
                       id='1024D/17B05A8F',
                       fingerprint='84D205F03E1E67096CB54E262BE83793AACCD97C',
                       size=1024, algorithm='D')
        key_two = dict(owner='foo bar',
                       id='1024D/17B06B90',
                       fingerprint='84D205F03E1E67096CB54E262BE8379317B06B90',
                       size=1024, algorithm='D')
        with config.set_config_value('gpghandler', 'enable_test_endpoint',
                                     True):
            fixture = self.useFixture(GPGAppFixture())
            resp = fixture.post_json(
                '/test/add_keys',
                {'keys': [key_one, key_two]})
            self.assertEqual(200, resp.status_code)
            self.assertTrue(
                database.key_exists(
                    fixture.db_connection, key_one['fingerprint']))
            self.assertTrue(
                database.key_exists(
                    fixture.db_connection, key_two['fingerprint']))

    def test_reset_db_returns_400_on_missing_key_data(self):
        with config.set_config_value('gpghandler', 'enable_test_endpoint',
                                     True):
            fixture = self.useFixture(GPGAppFixture())
            resp = fixture.post_json('/test/add_keys', dict(foo='bar'))
            self.assertThat(
                resp,
                IsAPIResponse(
                    400,
                    "Error: JSON Content-Type set but missing 'keys' key."))

    def test_returns_400_on_badly_formatted_key_data(self):
        fingerprint = '84D205F03E1E67096CB54E262BE8379317B06B90'
        with config.set_config_value('gpghandler', 'enable_test_endpoint',
                                     True):
            fixture = self.useFixture(GPGAppFixture())
            resp = fixture.post_json(
                '/test/add_keys',
                {'keys': [dict(ownder='myrandomid',
                               ib='1024D/17B05A8F',
                               figgerprint=fingerprint,
                               siz=1024, gorithm='D')]})
            self.assertThat(resp, IsAPIResponse(400, "Error: Bad Key format."))


class NagiosTestCase(TestCase):

    def test_nagios_postgres_correctly(self):
        self.fixture = self.useFixture(GPGAppFixture())
        resp = self.fixture.client.get('/nagios/postgres')
        self.assertEqual(200, resp.status_code, resp.data)
        self.assertEqual('application/json', resp.content_type)
        data = json.loads(resp.data)
        expected = {'ok': True}
        self.assertEqual(expected, data)

    def create_role(self, user, password):
        user = sql.expression.bindparam('user', AsIs(user))
        query = ("CREATE ROLE :user PASSWORD :password "
                 "NOSUPERUSER NOCREATEDB NOCREATEROLE INHERIT LOGIN")
        database.engine.execute(sql.text(query).bindparams(
            user, password=password))
        query = "DROP ROLE :user"
        self.addCleanup(database.engine.execute,
                        sql.text(query).bindparams(user))

    @skipIf(config.config.get('database', 'type') != 'postgres',
            "Roles aren't supported in sqlite")
    def test_nagios_postgres_misconfigure_role(self):
        self.create_role('notvaliduser', '3feab2')
        with config.set_config_value('database', 'user', 'notvaliduser'), \
                config.set_config_value('database', 'password', '3feab2'):
            self.fixture = self.useFixture(
                GPGAppFixture(recreate_engine=True))

            resp = self.fixture.client.get('/nagios/postgres')
            self.assertEqual(500, resp.status_code)
            self.assertIn("permission denied for relation",
                          resp.data)


class IsAPIResponse(Matcher):

    def __init__(self, expected_code, expected_status):
        super(IsAPIResponse, self).__init__()
        self.expected_code = expected_code
        self.expected_status = expected_status

    def match(self, matchee):
        if matchee.status_code != self.expected_code:
            return Mismatch("Response status code was %d, expecting %d" % (
                matchee.status_code, self.expected_code))
        if matchee.content_type != 'application/json':
            return Mismatch(
                "Response content_type was '%s', "
                "expected 'application/json'" % (matchee.content_type))
        data = json.loads(matchee.data)
        if data['status'] != self.expected_status:
            return Mismatch("Response status was '%s', expected '%s'" % (
                data['status'], self.expected_status))

    def __str__(self):
        return "%s(%r, %r)" % (
            self.__class__.__name__, self.expected_code, self.expected_status)


def test_suite():
    from unittest import TestLoader
    return TestLoader().loadTestsFromName(__name__)
