# Copyright 2016 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Web API functions."""
import base64

from flask import (
    g,
    jsonify,
    request,
)

import gpgme

from gpgservice import (
    database,
    errors,
    handler,
    validators,
    )


class APIError(Exception):

    """An exception that, when raised, will result in a json document and
    appropriate return code. Raise subclasses of this class.
    """

    status_code = 400

    def __init__(self, message, status_code=None, payload=None):
        super(APIError, self).__init__()
        self.message = message
        if status_code is not None:
            self.status_code = status_code
        self.payload = payload

    def to_dict(self):
        rv = dict(self.payload or ())
        rv['status'] = self.message
        return rv


def list_user_keys(user_id):
    user_id = decode_userid_or_raise(user_id)
    keys = database.get_keys_for_owner(g.database, user_id)
    display_keys = [db_to_presentation_format(k) for k in keys]
    return jsonify(user=user_id, keys=display_keys)


def add_user_fingerprint(user_id):
    user_id = decode_userid_or_raise(user_id)
    if request.json is None:
        raise APIError("Error: Missing JSON document in request body.")
    try:
        fingerprint = request.json['fingerprint']
    except KeyError:
        raise APIError("Error: JSON document missing 'fingerprint' key.")
    sane_fingerprint = handler.handler_singleton.sanitizeFingerprint(
        fingerprint)
    if sane_fingerprint is None:
        raise APIError("Error: Invalid fingerprint '%s'." % fingerprint)

    # This could be a new key, or it could be a user re-activating a key
    # they previously deactivated. Check for deactivated keys first:
    db = g.database
    key_to_return = None
    return_status = 201
    existing_key = database.get_keys_by_fingerprints(db, sane_fingerprint)
    if existing_key:
        existing_key = existing_key[0]
        if existing_key.owner == user_id:
            key_to_return = existing_key
            if existing_key.enabled:
                return_status = 200
            else:
                database.enable_key(db, sane_fingerprint)
        else:
            raise APIError("Error: Fingerprint already in database.", 409)
    else:  # Key didn't exist, add a new one:
        try:
            key = handler.handler_singleton.retrieveKey(sane_fingerprint)
            algorithm = handler.GPGKeyAlgorithm.items[key.algorithm].name
            gpg_key = database.GPGKey(
                owner=user_id, keyid=key.keyid, fingerprint=key.fingerprint,
                keysize=key.keysize, algorithm=algorithm, enabled=True,
                can_encrypt=key.can_encrypt)
            database.insert_key(db, gpg_key)
            key_to_return = gpg_key
        except errors.GPGKeyTemporarilyNotFoundError:
            raise APIError("Error: Keyserver unavailable.", 503)
        except errors.GPGKeyDoesNotExistOnServer:
            raise APIError("Error: Fingerprint not found on keyserver.", 404)
        # Even thought we check the databse for the key's existance above,
        # this can still happen if we have more than one frontend processing
        # requests.
        except database.DuplicateKeyInsertionError:
            raise APIError("Error: Fingerprint already in database.", 409)
    return jsonify(**db_to_presentation_format(key_to_return)), return_status


def disable_user_key(user_id, fingerprint):
    user_id = decode_userid_or_raise(user_id)
    validate_fingerprint_or_raise(fingerprint)
    if not database.owner_owns_fingerprint(g.database,
                                           user_id, fingerprint):
        raise APIError("Error: Fingerprint not found.", 404)
    if not database.disable_key(g.database,
                                fingerprint):
        # This might happen if we have multiple frontend workers, and
        # someone issues simultanious activate & deactivate requests.
        # It is uncovered by test code, since it's not worth getting into
        # mocking database calls to simulate this rare occurance. The
        # database call itself is tested, however.
        raise APIError("Error: Fingerprint not found.", 404)
    return jsonify(status='Key disabled')


def get_keys_by_fingerprint(fingerprint):
    if ',' in fingerprint:
        fingerprints = fingerprint.split(',')
        (validate_fingerprint_or_raise(f) for f in fingerprints)
        keys = database.get_keys_by_fingerprints(g.database, fingerprints)
        return jsonify(
            dict(keys=[db_to_presentation_format(k)
                 for k in keys]))
    else:
        validate_fingerprint_or_raise(fingerprint)
        keys = database.get_keys_by_fingerprints(g.database, fingerprint)
        if not keys:
            raise APIError("Error: Fingerprint not found.", 404)
        return jsonify(db_to_presentation_format(keys[0]))


def verify_content():
    signature = document = None
    for _, file_storage in request.files.items():
        if file_storage.content_type == 'application/pgp-signature':
            signature = file_storage.read()
        else:
            document = file_storage.read()

    if document is None:
        raise APIError("Error: Missing document to verify.")
    try:
        sig = handler.handler_singleton.getVerifiedSignature(
            document, signature)
        headers = {'X-GPG-Verified-Fingerprint': sig.fingerprint}
        return sig.plain_data, 200, headers
    except errors.GPGVerificationError as e:
        if getattr(e, 'code', None) == gpgme.ERR_NO_DATA:
            raise APIError("Error: Could not find any GPG verified content.")
        elif getattr(e, 'code', None) == gpgme.ERR_NO_PUBKEY:
            raise APIError("Error: Could not find GPG public key.")
        else:
            raise APIError("Error: %s." % e.message)
    except errors.GPGKeyNotFoundError as e:
        raise APIError("Error: %s." % e.message)


def encrypt_content():
    if 'X-GPG-Fingerprint' not in request.headers:
        raise APIError('Error: Missing X-GPG-Fingerprint header.')
    fingerprint = request.headers['X-GPG-Fingerprint']
    sane_fingerprint = handler.handler_singleton.sanitizeFingerprint(
        fingerprint)
    if sane_fingerprint is None:
        raise APIError("Error: Invalid fingerprint '%s'." % fingerprint)
    content = request.data
    if not content:
        raise APIError("Error: No content to encrypt.")
    try:
        key = handler.handler_singleton.retrieveActiveKey(sane_fingerprint)
        cipher = handler.handler_singleton.encryptContent(content,
                                                          key.fingerprint)
        if cipher is None:
            raise APIError("Error: Could not retrieve key '%s'." % fingerprint)
        return cipher, 200  # TODO, is this correct?
    except errors.GPGKeyRevoked:
        raise APIError("Error: GPG key '%s' has been revoked." % fingerprint)
    except errors.GPGKeyExpired:
        raise APIError("Error: GPG key '%s' has expired." % fingerprint)
    except errors.GPGKeyNotFoundError:
        raise APIError("Error: Could not retrieve key '%s'." % fingerprint)
    except TypeError:
        raise APIError("Error: Content can not be unicode")  # TODO
    except errors.GPGKeyCannotEncrypt:
        raise APIError("Error: GPG key '%s' cannot be used to encrypt."
                       % fingerprint)


def nagios_postgres():
    # return 200 or crash
    # check PG
    try:
        database.key_exists(g.database, 'not-a-fingerprint')
    except Exception as e:
        raise APIError(
            "Error: While checking database connection: %s" % e, 500)
    return jsonify(ok=True)


def test_reset_database():
    with g.database.begin():
        g.database.execute(database.GPGKeyTable.delete())
        handler.handler_singleton.resetLocalState()
        if request.json is not None:
            test_add_keys()
    return "OK", 200


def test_add_keys():
    try:
        keys_data = request.json['keys']
    except KeyError:
        raise APIError(
            "Error: JSON Content-Type set but missing 'keys' key.")
    else:
        try:
            for k in keys_data:
                database.insert_key(g.database,
                                    presentation_to_db_format(k))
        except KeyError:
            raise APIError("Error: Bad Key format.")
    return "OK", 200


def get_key_from_keyserver(fingerprint):
    validate_fingerprint_or_raise(fingerprint)
    try:
        key = handler.handler_singleton.retrieveKey(fingerprint)
        # 'key' is a PymeKey, rather than a GPGKey, so we can't use
        # the presentation function below. If we ever need to export this
        # type elsewhere, this should be extracted into a separate function:
        attrs_to_export = (
            'fingerprint', 'algorithm', 'revoked', 'expired', 'secret',
            'keysize', 'keyid', 'emails', 'displayname', 'owner_trust',
            'can_encrypt', 'can_sign', 'can_certify', 'can_authenticate')
        return jsonify(**{k: getattr(key, k) for k in attrs_to_export})
    except handler.GPGKeyNotFoundError:
        raise APIError(
            "Error: Key '%s' not found on keyserver." % fingerprint,
            404)


# Utility methods below here, API methods above ^^.
def decode_userid_or_raise(user_id):
    try:
        decoded_user_id = base64.b64decode(
            user_id.encode('utf8'), altchars='-_').decode('utf8')
        if not validators.valid_userid(decoded_user_id):
            raise APIError("Error: Invalid user id.", 400)
        return decoded_user_id
    except TypeError:
        raise APIError("Error: Invalid user id.", 400)


def validate_fingerprint_or_raise(fingerprint):
    if not validators.valid_fingerprint(fingerprint):
        raise APIError("Error: Invalid fingerprint.", 400)


def db_to_presentation_format(k):
    return {
        'id': k.keyid,
        'fingerprint': k.fingerprint,
        'size': k.keysize,
        'algorithm': k.algorithm,
        'enabled': k.enabled,
        'can_encrypt': k.can_encrypt,
        'owner': k.owner,
    }


def presentation_to_db_format(k):
    can_encrypt = k.get('can_encrypt', False)
    enabled = k.get('enabled', True)
    key = database.GPGKey(owner=k['owner'], keyid=k['id'],
                          fingerprint=k['fingerprint'], keysize=k['size'],
                          algorithm=k['algorithm'], enabled=enabled,
                          can_encrypt=can_encrypt)
    return key
