# Copyright 2016 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

import os.path
import uuid

import flask

from gpgservice import (
    config,
    database,
    logging,
    webapi,
    )


def create_flask_application(supress_logging=False):
    version_info = get_version_info()
    if not supress_logging:
        logging.configure_logging(config.config, 'gpgservice',
                                  revision=version_info)

    app = flask.Flask('gpgservice')

    app.add_url_rule('/', 'home',
                     lambda: 'gpgservice - Copyright 2016 Canonical')
    app.add_url_rule('/users/<user_id>/keys', 'list_user_keys',
                     webapi.list_user_keys)
    app.add_url_rule('/users/<user_id>/keys', 'add_user_fingerprint',
                     webapi.add_user_fingerprint, methods=['POST'])
    app.add_url_rule('/users/<user_id>/keys/<fingerprint>',
                     'disable_user_key', webapi.disable_user_key,
                     methods=['DELETE'])
    app.add_url_rule('/keys/<fingerprint>', 'get_keys_by_fingerprint',
                     webapi.get_keys_by_fingerprint)
    app.add_url_rule('/keyserver/<fingerprint>', 'get_key_from_keyserver',
                     webapi.get_key_from_keyserver)
    app.add_url_rule('/verify', 'verify_content', webapi.verify_content,
                     methods=['POST'])
    app.add_url_rule('/encrypt', 'encrypt_content', webapi.encrypt_content,
                     methods=['POST'])
    app.add_url_rule('/nagios/postgres',
                     'nagios_postgres', webapi.nagios_postgres)
    if config.config.getboolean('gpghandler', 'enable_test_endpoint'):
        app.add_url_rule('/test/reset_db', 'test_reset_database',
                         webapi.test_reset_database, methods=['POST'])
        app.add_url_rule('/test/add_keys', 'test_add_keys',
                         webapi.test_add_keys, methods=['POST'])

    @app.errorhandler(webapi.APIError)
    def handle_api_error(error):
        response = flask.jsonify(error.to_dict())
        response.status_code = error.status_code
        return response

    @app.before_request
    def before(logger=app.logger):
        request_id = flask.request.headers.get('X-Request-ID')
        if request_id is None:
            request_id = str(uuid.uuid1())
        logger.add_extra_args({'request_id': request_id})
        flask._request_ctx_stack.top.request_id = request_id
        flask.g.database = database.engine.connect()

    @app.after_request
    def after(response, logger=app.logger):
        if version_info is not None:
            response.headers['X-GPGService-Revision'] = version_info
        response.headers['X-Request-Id'] = \
            flask._request_ctx_stack.top.request_id
        flask.g.database.close()
        return response

    @app.teardown_request
    def teardown(exception, logger=app.logger):
        """Work after everything is really finished.

        This function must never fail.
        """
        try:
            # reset extra args in the logger
            logger.reset_extra_args()
        except Exception as err:
            logger.exception("Problem at request teardown: %s", err)

    app.config['MAX_CONTENT_LENGTH'] = config.config.getint(
        'gpghandler', 'maximum_upload_size')

    if not supress_logging:
        app.logger.info("Service started (r=%s)", version_info)
    return app


def get_version_info():
    path = os.path.join(os.path.dirname(__file__), '..', 'version-info.txt')
    try:
        with open(path, 'r') as version_info_file:
            return version_info_file.read().strip()
    except IOError:
        return None


app = create_flask_application()
